# image.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import cairo, random, math
from gi.repository import Gtk, Gdk, Gio, GdkPixbuf, Pango, GLib
from .history_manager import DrHistoryManager
from .selection_manager import DrSelectionManager
from .properties import DrPropertiesDialog
from .utilities_files import InvalidFileFormatException
from .utilities_overlay import utilities_generic_canvas_outline

class DrMotionBehavior():
	_LIMIT = 10

	HOVER = 0
	DRAW = 1
	SLIP = 2

class NoPixbufNoChangeException(Exception):
	def __init__(self, pb_name):
		# Context: an error message
		message = _("New pixbuf empty, no change applied to %s")
		super().__init__(message % pb_name)

################################################################################

@Gtk.Template(resource_path='/com/github/maoschanz/drawing/ui/image.ui')
class DrImage(Gtk.Box):
	__gtype_name__ = 'DrImage'

	_drawing_area = Gtk.Template.Child()
	_h_scrollbar = Gtk.Template.Child()
	_v_scrollbar = Gtk.Template.Child()
	reload_info_bar = Gtk.Template.Child()
	reload_label = Gtk.Template.Child()

	# HiDPI scale factor
	SCALE_FACTOR = 1.0 # XXX doesn't work well enough to be anything else

	# Threshold between normal rendering and crisp rendering
	ZOOM_THRESHOLD = 4.0

	# Maximal level of zoom (crisp rendering only)
	ZOOM_MAX = 2000

	def __init__(self, window, **kwargs):
		super().__init__(**kwargs)
		self.window = window

		self.gfile = None
		self.filename = None
		self._waiting_for_monitor = False
		self._gfile_monitor = None
		self._can_reload()

		# Closing the info bar
		self.reload_info_bar.connect('close', self.hide_reload_message)
		self.reload_info_bar.connect('response', self.hide_reload_message)

		# Framerate limit
		self._rendering_is_locked = False
		self._framerate_hint = 0

		self._ctrl_pressed = False

		if self.window.devel_mode:
			# Framerate tracking (debug only)
			self._skipped_frames = 0
			self._fps_counter = 0
			if self.window.should_track_framerate:
				self.reset_fps_counter()

		self._init_drawing_area()

		self._update_background_color()
		self.window.gsettings.connect('changed::ui-background-rgba', \
		                                          self._update_background_color)

		self._update_zoom_behavior()
		self.window.gsettings.connect('changed::ctrl-zoom', \
		                                             self._update_zoom_behavior)

	def _init_drawing_area(self):
		self._drawing_area.add_events( \
			Gdk.EventMask.BUTTON_PRESS_MASK | \
			Gdk.EventMask.BUTTON_RELEASE_MASK | \
			Gdk.EventMask.POINTER_MOTION_MASK | \
			Gdk.EventMask.SMOOTH_SCROLL_MASK | \
			Gdk.EventMask.ENTER_NOTIFY_MASK | \
			Gdk.EventMask.LEAVE_NOTIFY_MASK)
		# Using BUTTON_MOTION_MASK instead of POINTER_MOTION_MASK would be less
		# algorithmically complex but not "powerful" enough.

		# For displaying things on the widget
		self._drawing_area.connect('draw', self.on_draw)

		# For drawing with tools
		self._drawing_area.connect('motion-notify-event', self.on_motion_on_area)
		self._drawing_area.connect('button-press-event', self.on_press_on_area)
		self._drawing_area.connect('button-release-event', self.on_release_on_area)

		# For scrolling
		self._drawing_area.connect('scroll-event', self.on_scroll_on_area)
		self._h_scrollbar.connect('value-changed', self.on_scrollbar_value_change)
		self._v_scrollbar.connect('value-changed', self.on_scrollbar_value_change)

		# For the cursor
		self._drawing_area.connect('enter-notify-event', self.on_enter_image)
		self._drawing_area.connect('leave-notify-event', self.on_leave_image)

	def _update_background_color(self, *args):
		rgba = self.window.gsettings.get_strv('ui-background-rgba')
		self._bg_rgba = (float(rgba[0]), float(rgba[1]), \
		                                         float(rgba[2]), float(rgba[3]))
		# We remember this data here for performance: it will eb used by the
		# `on_draw` method which is called a lot, and reading a gsettings costs
		# a lot.

	def _update_zoom_behavior(self, *args):
		self._ctrl_to_zoom = self.window.gsettings.get_boolean('ctrl-zoom')

	############################################################################
	# Image initialization #####################################################

	def _init_image_common(self):
		"""Part of the initialization common to both a new blank image and an
		opened image."""
		self._is_pressed = False

		# Zoom and scroll initialization
		self.scroll_x = 0
		self.scroll_y = 0
		self.zoom_level = 1.0
		self.motion_behavior = DrMotionBehavior.HOVER
		self._slip_press_x = 0.0
		self._slip_press_y = 0.0
		self._slip_init_x = 0.0
		self._slip_init_y = 0.0

		# Selection initialization
		self.selection = DrSelectionManager(self)

		# History initialization
		self._history = DrHistoryManager(self)
		self.set_action_sensitivity('undo', False)
		self.set_action_sensitivity('redo', False)

	def init_background(self, width, height, background_rgba):
		self._init_image_common()
		self._history.set_initial_operation(background_rgba, None, width, height)
		self.restore_last_state()

	def try_load_pixbuf(self, pixbuf):
		self._init_image_common()
		self._load_pixbuf_common(pixbuf)
		self.restore_last_state()
		self.update_title()

	def _new_blank_pixbuf(self, w, h):
		return GdkPixbuf.Pixbuf.new(GdkPixbuf.Colorspace.RGB, True, 8, w, h)

	def restore_last_state(self):
		"""Set the last saved pixbuf from the history as the main_pixbuf. This
		is used to rebuild the picture from its history."""
		last_saved_pixbuf_op = self._history.get_last_saved_state()
		self._apply_state(last_saved_pixbuf_op)

	def reset_to_initial_pixbuf(self):
		self._apply_state(self._history.initial_operation)
		self._history.rewind_history()

	def _apply_state(self, state_op):
		# restore the state found in the history
		pixbuf = state_op['pixbuf']
		width = state_op['width']
		height = state_op['height']
		self.set_temp_pixbuf(self._new_blank_pixbuf(1, 1))
		self.selection.init_pixbuf()
		self.surface = cairo.ImageSurface(cairo.Format.ARGB32, width, height)
		if pixbuf is None:
			# no pixbuf in the operation: the restored state is a blank one
			rgba = state_op['rgba']
			r = rgba.red
			g = rgba.green
			b = rgba.blue
			a = rgba.alpha
			self.set_main_pixbuf(self._new_blank_pixbuf(width, height))
			cairo_context = cairo.Context(self.surface)
			cairo_context.set_source_rgba(r, g, b, a)
			cairo_context.paint()
			self.update()
			self.set_surface_as_stable_pixbuf()
		else:
			self.set_main_pixbuf(state_op['pixbuf'].copy())
			self.use_stable_pixbuf()

	############################################################################
	# (re)loading the pixbuf of a given file ###################################

	def _load_pixbuf_common(self, pixbuf):
		if not pixbuf.get_has_alpha():
			pixbuf = pixbuf.add_alpha(False, 255, 255, 255)
		background_rgba = self.window.gsettings.get_strv('default-rgba')
		self._history.set_initial_operation(background_rgba, pixbuf, \
		                                pixbuf.get_width(), pixbuf.get_height())
		self.set_main_pixbuf(pixbuf)

	def reload_from_disk(self):
		"""Safely reloads the image from the disk."""
		if self.gfile is None:
			# the action shouldn't be active in the first place
			return
		disk_pixbuf = GdkPixbuf.Pixbuf.new_from_file(self.get_file_path())
		self._load_pixbuf_common(disk_pixbuf)
		self.window.update_picture_title()
		self.use_stable_pixbuf()
		self.update()
		self.remember_current_state()

	def try_load_file(self, gfile):
		try:
			self.gfile = gfile
			pixbuf = GdkPixbuf.Pixbuf.new_from_file(self.get_file_path())
			self.connect_gfile_monitoring()
		except Exception as ex:
			if not ex.message:
				ex.message = "[exception without a valid message]"
			ex = InvalidFileFormatException(ex.message, gfile.get_path())
			self.window.reveal_action_report(ex.message)
			self.gfile = None
			pixbuf = self._new_blank_pixbuf(100, 100)
			# XXX dans l'idéal on devrait ne rien ouvrir non ? ou si besoin (si
			# ya pas de fenêtre) ouvrir un truc respectant les settings, plutôt
			# qu'un petit pixbuf corrompu
		self.try_load_pixbuf(pixbuf)
		self._can_reload()

	def connect_gfile_monitoring(self):
		flags = Gio.FileMonitorFlags.WATCH_MOUNTS
		self._gfile_monitor = self.gfile.monitor(flags)
		self._gfile_monitor.connect('changed', self.reveal_reload_message)

	def set_monitoring(self, value):
		self._waiting_for_monitor = value

	def reveal_reload_message(self, *args):
		if self._waiting_for_monitor:
			# I'm not sure this lock is 100% correct because i'm monitoring the
			# portal proxy file when testing with flatpak. Better than nothing.
			if args[3] != Gio.FileMonitorEvent.CHANGED:
				self._waiting_for_monitor = False
			return
		self._can_reload()
		self.reload_label.set_visible(self.window.get_allocated_width() > 500)
		self.reload_info_bar.set_visible(True)

	def hide_reload_message(self, *args):
		self.reload_info_bar.set_visible(False)

	def _can_reload(self):
		self.set_action_sensitivity('reload_file', self.gfile is not None)

	############################################################################
	# Image title and tab management ###########################################

	def build_tab_widget(self):
		"""Build the GTK widget displayed as the tab title."""
		# The tab can be closed with a button.
		btn = Gtk.Button.new_from_icon_name('window-close-symbolic', Gtk.IconSize.BUTTON)
		btn.set_relief(Gtk.ReliefStyle.NONE)
		btn.connect('clicked', self.try_close_tab)
		# The title is a label. Middle-clicking on it closes the tab too.
		self.tab_label = Gtk.Label(label=self.get_filename_for_display())
		self.tab_label.set_ellipsize(Pango.EllipsizeMode.END)
		event_box = Gtk.EventBox()
		event_box.add(self.tab_label)
		event_box.connect('button-press-event', self.on_tab_title_clicked)
		# These widgets are packed in a regular box, which is returned.
		tab_title = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, expand=True)
		if self.window.deco_layout == 'he':
			tab_title.pack_start(btn, expand=False, fill=False, padding=0)
			tab_title.pack_end(event_box, expand=True, fill=True, padding=0)
		else:
			tab_title.pack_start(event_box, expand=True, fill=True, padding=0)
			tab_title.pack_end(btn, expand=False, fill=False, padding=0)
		tab_title.show_all()
		return tab_title
	
	def on_tab_title_clicked(self, widget, event_button):
		if event_button.type == Gdk.EventType.BUTTON_PRESS \
		and event_button.button == Gdk.BUTTON_MIDDLE:
			self.try_close_tab()
			return True
		return False # This callback HAS TO return a boolean

	def update_title(self):
		main_title = self.get_filename_for_display()
		if not self.is_saved():
			main_title = "*" + main_title
		self.set_tab_label(main_title)
		return main_title

	def get_filename_for_display(self):
		if self.get_file_path() is None:
			unsaved_file_name = _("Unsaved file")
		else:
			unsaved_file_name = self.get_file_path().split('/')[-1]
		return unsaved_file_name

	def try_close_tab(self, *args):
		"""Ask the window to close the image/tab. Then unallocate widgets and
		pixbufs."""
		if self.window.close_tab(self):
			self.destroy()
			self.selection.reset(False)
			self.main_pixbuf = None
			self.temp_pixbuf = None
			self._history.empty_history()
			return True
		else:
			return False

	def set_tab_label(self, title_text):
		self.tab_label.set_label(title_text)

	def get_file_path(self):
		if self.gfile is None:
			return None
		else:
			return self.gfile.get_path()

	def show_properties(self):
		DrPropertiesDialog(self.window, self)

	def update_image_wide_actions(self):
		self.update_history_sensitivity()
		self._can_reload()

	############################################################################
	# History management #######################################################

	def try_undo(self):
		self._history.try_undo()

	def try_redo(self):
		self._history.try_redo()

	def is_saved(self):
		return self._history.get_saved()

	def remember_current_state(self):
		self._history.add_state(self.main_pixbuf.copy())

	def update_history_sensitivity(self):
		self.set_action_sensitivity('undo', self._history.can_undo())
		self.set_action_sensitivity('redo', self._history.can_redo())
		# self.update_history_actions_labels()

	def add_to_history(self, operation):
		self._history.add_operation(operation)

	def should_replace(self):
		if self._history.can_undo():
			return False
		return not self._history.has_initial_pixbuf()

	def get_initial_rgba(self):
		return self._history.initial_operation['rgba']

	############################################################################
	# Misc ? ###################################################################

	def set_action_sensitivity(self, action_name, state):
		self.window.lookup_action(action_name).set_enabled(state)

	def update_actions_state(self):
		# XXX shouldn't it be done by the selection_manager?
		state = self.selection.is_active
		self.set_action_sensitivity('unselect', state)
		self.set_action_sensitivity('select_all', not state)
		self.set_action_sensitivity('selection_cut', state)
		self.set_action_sensitivity('selection_copy', state)
		self.set_action_sensitivity('selection_delete', state)
		self.set_action_sensitivity('selection_export', state)
		self.set_action_sensitivity('new_tab_selection', state)
		self.set_action_sensitivity('selection-replace-canvas', state)
		self.set_action_sensitivity('selection-expand-canvas', state)
		self.active_tool().update_actions_state()

	def active_tool(self):
		return self.window.active_tool()

	def get_mouse_is_pressed(self):
		return self._is_pressed

	############################################################################
	# Drawing area, main pixbuf, and surface management ########################

	def on_draw(self, area, cairo_context):
		"""Signal callback. Executed when self._drawing_area is redrawn."""
		if self.window.devel_mode:
			self._fps_counter += 1

		# Background color
		cairo_context.set_source_rgba(*self._bg_rgba)
		cairo_context.paint()

		# Zoom level
		cairo_context.scale(self.zoom_level, self.zoom_level)

		# Image (with scroll position)
		cairo_context.set_source_surface(self.get_surface(), \
		                                 -1 * self.scroll_x, -1 * self.scroll_y)
		if self.is_zoomed_surface_sharp():
			cairo_context.get_source().set_filter(cairo.FILTER_NEAREST)
		cairo_context.paint()

		# What the tool shows on the canvas, upon what it paints, for example an
		# overlay to imply how to interact with a previewed operation.
		self.active_tool().on_draw_above(area, cairo_context)

		# Limit of the canvas (for readability)
		utilities_generic_canvas_outline(cairo_context, self.zoom_level, \
		                              self.get_pixbuf_width() - self.scroll_x, \
		                             self.get_pixbuf_height() - self.scroll_y)

	def on_press_on_area(self, area, event):
		"""Signal callback. Executed when a mouse button is pressed on
		self._drawing_area, if the button is the mouse wheel the colors are
		exchanged, otherwise the signal is transmitted to the selected tool."""
		if self._is_pressed:
			# reject attempts to draw with a given button if an other one is
			# already doing something
			return
		event_x, event_y = self.get_event_coords(event)
		if event.button == 2:
			self.motion_behavior = DrMotionBehavior.SLIP
			self._slip_press_x = event.x
			self._slip_press_y = event.y
			self._slip_init_x = self.scroll_x
			self._slip_init_y = self.scroll_y
			return
		self.motion_behavior = DrMotionBehavior.DRAW
		self._is_pressed = True
		self.window.set_window_subtitles()
		# subtitles must be generated *before* calling the tool, otherwise any
		# property changed by a modifier would be reset by `get_editing_tips`
		self.active_tool().on_press_on_area(event, self.surface, event_x, event_y)

	def on_motion_on_area(self, area, event):
		"""Signal callback. Executed when the mouse pointer moves upon
		self._drawing_area, the signal is transmitted to the selected tool.
		If a button (not the mouse wheel) is pressed, the tool's method should
		have an effect on the image, otherwise it shouldn't change anything
		except the mouse cursor icon for example."""
		event_x, event_y = self.get_event_coords(event)

		if self.motion_behavior == DrMotionBehavior.HOVER:
			# Some tools need the coords in the image, others need the coords on
			# the widget, so the entire event is given.
			self.active_tool().on_unclicked_motion_on_area(event, self.surface)

		elif self.motion_behavior == DrMotionBehavior.DRAW:
			# implicitly impossible if not self._is_pressed
			self.active_tool().on_motion_on_area(event, self.surface, event_x, \
			                                 event_y, self._rendering_is_locked)
			if self._rendering_is_locked:
				if self.window.devel_mode:
					self._skipped_frames += 1
				return
			self._rendering_is_locked = True
			self.update()
			GLib.timeout_add(self._framerate_hint, self._async_unlock, {})

		else: # self.motion_behavior == DrMotionBehavior.SLIP:
			self.scroll_x = self._slip_init_x
			self.scroll_y = self._slip_init_y
			delta_x = self._slip_press_x - event.x
			delta_y = self._slip_press_y - event.y
			self.add_deltas(delta_x, delta_y, 1 / self.zoom_level)

		# If <Ctrl> is pressed, a tooltip displaying contextual information is
		# shown: by default, it contains at least the pointer coordinates.
		if (event.state & Gdk.ModifierType.CONTROL_MASK) == Gdk.ModifierType.CONTROL_MASK:
			self._ctrl_pressed = True
		if self._ctrl_pressed:
			full_tooltip_text = str(event_x) + ", " + str(event_y)
			tool_tooltip = self._get_tool_tooltip(event_x, event_y)
			if tool_tooltip is not None:
				full_tooltip_text += "\n" + tool_tooltip
			self._drawing_area.set_tooltip_text(full_tooltip_text)
		else:
			self._drawing_area.set_tooltip_text(None)

	def on_release_on_area(self, area, event):
		"""Signal callback. Executed when a mouse button is released on
		self._drawing_area, if the button is not the signal is transmitted to
		the selected tool."""
		self._ctrl_pressed = False
		if self.motion_behavior == DrMotionBehavior.SLIP:
			if not self._is_slip_moving():
				self.window.on_middle_click()
			self.motion_behavior = DrMotionBehavior.HOVER
			return
		self.motion_behavior = DrMotionBehavior.HOVER
		event_x, event_y = self.get_event_coords(event)
		self.active_tool().on_release_on_area(event, self.surface, event_x, event_y)
		self._is_pressed = False
		self.window.update_picture_title() # just to add the star
		self.window.set_window_subtitles() # the tool's state changed

	def _is_slip_moving(self):
		"""Tells if the pointer moved while the middle button of the mouse is
		pressed, depending on a constant hardcoded limit."""
		mx = abs(self._slip_init_x - self.scroll_x) > DrMotionBehavior._LIMIT
		my = abs(self._slip_init_y - self.scroll_y) > DrMotionBehavior._LIMIT
		return mx or my

	def _get_tool_tooltip(self, ev_x, ev_y):
		return self.active_tool().get_tooltip(ev_x, ev_y ,self.motion_behavior)

	def update(self):
		# print('image.py: _drawing_area.queue_draw')
		self._drawing_area.queue_draw()

	def _async_unlock(self, content_params={}):
		"""This is used as a GSourceFunc so it should return False."""
		self._rendering_is_locked = False
		return False

	def get_surface(self):
		return self.surface

	def on_enter_image(self, *args):
		self.window.set_cursor(True)

	def on_leave_image(self, *args):
		self.window.set_cursor(False)

	def post_save(self):
		self.use_stable_pixbuf()
		self.update()

	def set_surface_as_stable_pixbuf(self):
		w = self.surface.get_width()
		h = self.surface.get_height()
		self.main_pixbuf = Gdk.pixbuf_get_from_surface(self.surface, 0, 0, w, h)
		self._framerate_hint = math.sqrt(w * h) - 1000
		self._framerate_hint = int(self._framerate_hint * 0.2)
		# between 500 and 33ms (= between 2 and 30 fps)
		self._framerate_hint = max(33, min(500, self._framerate_hint))
		# print("image.py: hint =", self._framerate_hint)

	def use_stable_pixbuf(self):
		"""This is called by tools' `restore_pixbuf`, so at the beginning of
		each operation (even unapplied)."""
		# maybe the "scale" parameter should be 1 instead of 0
		self.surface = Gdk.cairo_surface_create_from_pixbuf(self.main_pixbuf, 0, None)
		# print('image.py: use_stable_pixbuf')
		self.surface.set_device_scale(self.SCALE_FACTOR, self.SCALE_FACTOR)

	def get_pixbuf_width(self):
		return self.main_pixbuf.get_width()

	def get_pixbuf_height(self):
		return self.main_pixbuf.get_height()

	def set_main_pixbuf(self, new_pixbuf):
		"""Safely set a pixbuf as the main one (not used everywhere internally
		in image.py, but it's normal)."""
		if new_pixbuf is None:
			raise NoPixbufNoChangeException('main_pixbuf')
		else:
			self.main_pixbuf = new_pixbuf

	############################################################################
	# Temporary pixbuf management ##############################################

	def set_temp_pixbuf(self, new_pixbuf):
		if new_pixbuf is None:
			raise NoPixbufNoChangeException('temp_pixbuf')
		else:
			self.temp_pixbuf = new_pixbuf

	def reset_temp(self):
		self.set_temp_pixbuf(self._new_blank_pixbuf(1, 1))
		self.use_stable_pixbuf()
		self.update()

	############################################################################
	# Framerate tracking (debug only) ##########################################

	def reset_fps_counter(self, async_cb_data={}):
		"""Development only: live-display the evolution of the framerate of the
		drawing area. The max should be around 60, but many tools don't require
		so many redraws.
		This is used as a GSourceFunc so it should return False."""
		if self.window.should_track_framerate:
			# Context: this is a debug information that users will never see
			msg = _("%s frames per second") % self._fps_counter
			msg += " (" + str(self._skipped_frames) + " motion inputs skipped)"
			self.window.reveal_message(msg)
			self._fps_counter = 0
			self._skipped_frames = 0
			GLib.timeout_add(1000, self.reset_fps_counter, {})
		elif self.window.info_bar.get_visible():
			self.window.reveal_message("Tracking stopped.", True)
		return False

	############################################################################
	# Interaction with the minimap #############################################

	def get_widget_width(self):
		return self._drawing_area.get_allocated_width()

	def get_widget_height(self):
		return self._drawing_area.get_allocated_height()

	def generate_mini_pixbuf(self, preview_size):
		mpb_width = self.get_pixbuf_width()
		mpb_height = self.get_pixbuf_height()
		if mpb_height > mpb_width:
			mw = preview_size * (mpb_width/mpb_height)
			mh = preview_size
		else:
			mw = preview_size
			mh = preview_size * (mpb_height/mpb_width)
		return self.main_pixbuf.scale_simple(mw, mh, GdkPixbuf.InterpType.TILES)

	def get_minimap_need_overlay(self):
		mpb_width = self.get_pixbuf_width()
		mpb_height = self.get_pixbuf_height()
		show_x = self.get_widget_width() < mpb_width * self.zoom_level
		show_y = self.get_widget_height() < mpb_height * self.zoom_level
		show_overlay = show_x or show_y
		return show_overlay

	def get_minimap_ratio(self, mini_width):
		return mini_width/self.get_pixbuf_width()

	def get_visible_size(self):
		visible_width = int(self.get_widget_width() / self.zoom_level)
		visible_height = int(self.get_widget_height() / self.zoom_level)
		return visible_width, visible_height

	############################################################################
	# Scroll and zoom levels ###################################################

	def get_previewed_width(self):
		# Indirect way to know if it's a transform tool
		if self.active_tool().menu_id == 1:
			if not self.active_tool().apply_to_selection:
				return self.temp_pixbuf.get_width() + 12
		return self.get_pixbuf_width()

	def get_previewed_height(self):
		# Indirect way to know if it's a transform tool
		if self.active_tool().menu_id == 1:
			if not self.active_tool().apply_to_selection:
				return self.temp_pixbuf.get_height() + 12
		return self.get_pixbuf_height()

	def fake_scrollbar_update(self):
		self.add_deltas(0, 0, 0)

	def get_event_coords(self, event, as_integers=True):
		event_x = self.scroll_x + (event.x / self.zoom_level)
		event_y = self.scroll_y + (event.y / self.zoom_level)
		if as_integers:
			# `int()` will truncate to the lower integer so we need this to get
			# an accurate behavior when doing pixel-art for example
			event_x += 0.5
			event_y += 0.5
			return int(event_x), int(event_y)
		else:
			return event_x, event_y

	def get_corrected_coords(self, x1, x2, y1, y2, with_selection, with_zoom):
		"""Do whatever coordinates conversions are needed by tools like `crop`
		and `scale` to display things (selection pixbuf, mouse cursors on hover,
		etc.) correctly enough."""
		w = x2 - x1
		h = y2 - y1
		x1 = x1 - self.scroll_x
		y1 = y1 - self.scroll_y
		if with_selection:
			x1 += self.selection.selection_x
			y1 += self.selection.selection_y
		x2 = x1 + w
		y2 = y1 + h
		if with_zoom:
			x1 *= self.zoom_level
			x2 *= self.zoom_level
			y1 *= self.zoom_level
			y2 *= self.zoom_level
		return x1, x2, y1, y2

	def get_nineths_sizes(self, apply_to_selection, x1, y1):
		"""Returns the sizes of the 'nineths' of the image used for example by
		'scale' or 'crop' to decide the cursor they'll show."""
		height = self.temp_pixbuf.get_height()
		width = self.temp_pixbuf.get_width()
		if not apply_to_selection:
			x1 = 0
			y1 = 0
		# width_left, width_right, height_top, height_bottom
		wl, wr, ht, hb = self.get_corrected_coords(int(x1), width, int(y1), \
		                                       height, apply_to_selection, True)
		# XXX using local deltas this way "works" but isn't mathematically
		# correct: scaled selections have a "null" and excentred central ninth
		# ^ c'est toujours vrai ça ??
		wl += 0.4 * width * self.zoom_level
		wr -= 0.4 * width * self.zoom_level
		ht += 0.4 * height * self.zoom_level
		hb -= 0.4 * height * self.zoom_level
		return {'wl': wl, 'wr': wr, 'ht': ht, 'hb': hb}

	def on_scroll_on_area(self, area, event):
		# TODO https://lazka.github.io/pgi-docs/index.html#Gdk-3.0/classes/EventScroll.html#Gdk.EventScroll
		ctrl_is_used = (event.state & Gdk.ModifierType.CONTROL_MASK) == Gdk.ModifierType.CONTROL_MASK
		if ctrl_is_used == self._ctrl_to_zoom:
			self._zoom_to_point(event)
		else:
			acceleration = 20 / self._zoom_profile()
			self.add_deltas(event.delta_x, event.delta_y, acceleration)

	def on_scrollbar_value_change(self, scrollbar):
		self.correct_coords(self._h_scrollbar.get_value(), self._v_scrollbar.get_value())
		self.update() # allowing imperfect framerate would likely be useless

	def reset_deltas(self, delta_x, delta_y):
		if delta_x > 0:
			wanted_x = self.get_previewed_width()
		elif delta_x < 0:
			wanted_x = 0
		else:
			wanted_x = self.scroll_x

		if delta_y > 0:
			wanted_y = self.get_previewed_height()
		elif delta_y < 0:
			wanted_y = 0
		else:
			wanted_y = self.scroll_y

		self.correct_coords(wanted_x, wanted_y)
		self.window.minimap.update_overlay()

	def add_deltas(self, delta_x, delta_y, factor):
		wanted_x = self.scroll_x + int(delta_x * factor)
		wanted_y = self.scroll_y + int(delta_y * factor)
		self.correct_coords(wanted_x, wanted_y)
		self.window.minimap.update_overlay()

	def correct_coords(self, wanted_x, wanted_y):
		available_w = self.get_widget_width()
		available_h = self.get_widget_height()
		if available_w < 2:
			return # could be better handled

		# Update the horizontal scrollbar
		mpb_width = self.get_previewed_width()
		wanted_x = min(wanted_x, self.get_max_coord(mpb_width, available_w))
		wanted_x = max(wanted_x, 0)
		self.update_scrollbar(False, available_w, int(mpb_width), int(wanted_x))

		# Update the vertical scrollbar
		mpb_height = self.get_previewed_height()
		wanted_y = min(wanted_y, self.get_max_coord(mpb_height, available_h))
		wanted_y = max(wanted_y, 0)
		self.update_scrollbar(True, available_h, int(mpb_height), int(wanted_y))

	def get_max_coord(self, mpb_size, available_size):
		max_coord = mpb_size - (available_size / self.zoom_level)
		return max_coord

	def update_scrollbar(self, is_vertical, allocated_size, pixbuf_size, coord):
		if is_vertical:
			scrollbar = self._v_scrollbar
		else:
			scrollbar = self._h_scrollbar
		scrollbar.set_visible(allocated_size / self.zoom_level < pixbuf_size)
		scrollbar.set_range(0, pixbuf_size)
		scrollbar.get_adjustment().set_page_size(allocated_size / self.zoom_level)
		scrollbar.set_value(coord)
		if is_vertical:
			self.scroll_y = int(scrollbar.get_value())
		else:
			self.scroll_x = int(scrollbar.get_value())

	def _zoom_to_point(self, event):
		"""Zoom in or out the image in a way such that the point under the
		pointer stays as much as possible under the pointer."""
		event_x, event_y = self.get_event_coords(event)

		# Updating the zoom level
		zoom_delta = (event.delta_x + event.delta_y) * -1 * self._zoom_profile()
		self.inc_zoom_level(zoom_delta)

		new_event_x, new_event_y = self.get_event_coords(event)
		delta_correction_x = event_x - new_event_x
		delta_correction_y = event_y - new_event_y

		# Updating the scroll position based on the values previously found
		self.add_deltas(delta_correction_x, delta_correction_y, 1)

	def inc_zoom_level(self, delta):
		self.set_zoom_level((self.zoom_level * 100) + delta)

	def set_zoom_level(self, level):
		normalized_zoom_level = max(min(level, self.ZOOM_MAX), 20)
		self.zoom_level = (int(normalized_zoom_level)/100)
		self.window.minimap.update_zoom_scale(self.zoom_level)
		if self.is_zoomed_surface_sharp():
			self.window.minimap.set_zoom_label(self.zoom_level * 100)
		self.fake_scrollbar_update()
		self.update()

	def set_opti_zoom_level(self, *args):
		allocated_width = self.get_widget_width()
		allocated_height = self.get_widget_height()
		h_ratio = allocated_width / self.get_previewed_width()
		v_ratio = allocated_height / self.get_previewed_height()
		opti = min(h_ratio, v_ratio) * 99 # Not 100 because some margin is cool
		self.set_zoom_level(opti)
		self.scroll_x = 0
		self.scroll_y = 0

	def is_zoomed_surface_sharp(self):
		return self.zoom_level > self.ZOOM_THRESHOLD

	def _zoom_profile(self):
		"""This is the 'speed' of the zoom scrolling: when between 20% and 100%
		it's quite precise, but between 1000% and 2000% we prefer being as fast
		as possible."""
		if self.zoom_level < 2.0:
			return 3.0
		elif self.zoom_level < 4.0:
			return 6.0
		elif self.zoom_level < 10.0:
			return 10.0
		else:
			return 20.0

	############################################################################
################################################################################




# Licensed under GPL3 https://github.com/maoschanz/drawing/blob/master/LICENSE

import cairo

class AbstractBrush():
	__gtype_name__ = 'AbstractBrush'

	def __init__(self, brush_id, brush_tool, *args):
		self._id = brush_id
		self._tool = brush_tool

	def _get_tips(self, use_pressure, brush_direction):
		return [_("Brush")]

	############################################################################

	def draw_preview(self, operation, cairo_context):
		cairo_context.set_operator(operation['operator'])
		cairo_context.set_line_width(operation['line_width'])
		cairo_context.new_path()
		for pt in operation['path']:
			cairo_context.line_to(pt['x'], pt['y'])
		cairo_context.stroke()

	def operation_on_mask(self, operation, original_context):
		if operation['operator'] == cairo.Operator.CLEAR \
		or operation['operator'] == cairo.Operator.SOURCE:
			# When using CLEAR or SOURCE, we don't need to use a temporary
			# surface, and actually we can't because using it as a mask would
			# just erase the entire image.
			original_context.set_operator(operation['operator'])
			original_context.set_source_rgba(*operation['rgba'])
			self.do_masked_brush_op(original_context, operation)
			return

		# Creation of a blank surface with a new context; each brush decides how
		# to apply the options set by the user (`operation`), except for the
		# operator which has to be "wrongly" set to SOURCE.
		w = self._tool.get_surface().get_width()
		h = self._tool.get_surface().get_height()
		mask = cairo.ImageSurface(cairo.Format.ARGB32, w, h)
		context2 = cairo.Context(mask)

		if operation['antialias']:
			antialias = cairo.Antialias.DEFAULT
		else:
			antialias = cairo.Antialias.NONE
		context2.set_antialias(antialias)

		context2.set_operator(cairo.Operator.SOURCE)
		context2.set_source_rgba(*operation['rgba'])

		self.do_masked_brush_op(context2, operation)

		# Paint the surface onto the actual image with the chosen operator
		original_context.set_operator(operation['operator'])
		original_context.set_source_surface(mask)
		original_context.paint()

	############################################################################

	def do_brush_operation(self, cairo_context, operation):
		pass

	def do_masked_brush_op(self, cairo_context, operation):
		pass

	############################################################################
################################################################################




# select_color.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .abstract_select import AbstractSelectionTool
from .utilities_colors import utilities_get_rgba_name, \
                              utilities_gdk_rgba_from_xy, \
                              utilities_gdk_rgba_to_hexadecimal
from .utilities_paths import utilities_get_magic_path

class ToolColorSelect(AbstractSelectionTool):
	__gtype_name__ = 'ToolColorSelect'

	def __init__(self, window, **kwargs):
		# Context: this is a tool to "magically" select an area depending on its
		# color. For example clicking on a white pixel will select the
		# surrounding area made of white pixels.
		super().__init__('color_select', _("Color selection"), 'tool-magic-symbolic', window)

	def get_tooltip(self, event_x, event_y, motion_behavior):
		color = utilities_gdk_rgba_from_xy(self.get_surface(), event_x, event_y)
		if color is None:
			return None
		color_name = utilities_get_rgba_name(color)
		color_code = utilities_gdk_rgba_to_hexadecimal(color)
		return color_name + "\n" + color_code

	def get_editing_tips(self):
		tips = super().get_editing_tips()
		if not self.selection_is_active():
			label_warning1 = self.label + " - " + _("May not work for complex shapes")
			tips.append(label_warning1)
			label_warning2 = self.label + " - " + _("It will not work well " + \
				                               "if the area's edges are blurry")
			tips.append(label_warning2)
		return tips

	############################################################################

	def press_define(self, event_x, event_y):
		pass

	def motion_define(self, event_x, event_y, render):
		pass

	def release_define(self, surfc, event_x, event_y):
		path = utilities_get_magic_path(surfc, event_x, event_y, self.window, 1)
		self._pre_load_path(path)
		if path is None:
			return
		self.operation_type = 'op-define'
		operation = self.build_operation()
		self.apply_operation(operation)

	############################################################################
################################################################################




# Licensed under GPL3 https://github.com/maoschanz/drawing/blob/master/LICENSE

import cairo
from gi.repository import Gdk
from .abstract_filter import AbstractFilter

class FilterContrast(AbstractFilter):
	__gtype_name__ = 'FilterContrast'

	def __init__(self, filter_id, filters_tool, *args):
		super().__init__(filter_id, filters_tool)
		self._label, self._spinbtn = self._tool.bar.add_spinbtn( \
		                  _("Increase contrast"), [0, 0, 100, 5, 10, 0], 3, '%')
		# it's [value, lower, upper, step_increment, page_increment, page_size]

	def get_preferred_minimum_width(self):
		return self._label.get_preferred_width()[0] + \
		     self._spinbtn.get_preferred_width()[0]

	def set_filter_compact(self, is_active, is_compact):
		self._label.set_visible(is_active and not is_compact)
		self._spinbtn.set_visible(is_active)

	def build_filter_op(self):
		options = {
			'percent': self._spinbtn.get_value() / 100
		}
		return options

	def do_filter_operation(self, source_pixbuf, operation):
		"""Create a temp_pixbuf from a surface of the same size, whose cairo
		context is first painted using the original surface (source operator),
		which is basically a stupid way to copy it, and then painted again (with
		alpha this time) using a blending mode that will increase the contrast.
		Both OVERLAY, SOFT_LIGHT, and HARD_LIGHT can work as operators."""
		percent = operation['percent']
		surface = Gdk.cairo_surface_create_from_pixbuf(source_pixbuf, 0, None)
		scale = self._tool.scale_factor()
		surface.set_device_scale(scale, scale)
		width = source_pixbuf.get_width()
		height = source_pixbuf.get_height()
		new_surface = cairo.ImageSurface(cairo.Format.ARGB32, width, height)
		cairo_context = cairo.Context(new_surface)
		cairo_context.set_source_surface(surface)

		cairo_context.set_operator(cairo.Operator.SOURCE)
		cairo_context.paint()

		cairo_context.set_operator(cairo.Operator.SOFT_LIGHT)
		# OVERLAY SOFT_LIGHT HARD_LIGHT
		cairo_context.paint_with_alpha(percent)

		new_pixbuf = Gdk.pixbuf_get_from_surface(new_surface, 0, 0, \
		                      new_surface.get_width(), new_surface.get_height())
		self._tool.get_image().set_temp_pixbuf(new_pixbuf)

	############################################################################
################################################################################




# Licensed under GPL3 https://github.com/maoschanz/drawing/blob/master/LICENSE

from gi.repository import Gdk

from .utilities_paths import utilities_get_rgba_for_xy

################################################################################

def utilities_gdk_rgba_to_color_array(gdk_rgba):
	"""Convert a Gdk.RGBA to an array of values intended to be used directly by
	cairo. Thus the value of alpha between 0 and 1, despite the 3 other values
	being integers between 0 and 255."""
	r = int(255 * gdk_rgba.red)
	g = int(255 * gdk_rgba.green)
	b = int(255 * gdk_rgba.blue)
	return [r, g, b, gdk_rgba.alpha]

def utilities_gdk_rgba_to_normalized_array(gdk_rgba):
	"""Convert a Gdk.RGBA to an array of values between 0 and 1."""
	return [gdk_rgba.red, gdk_rgba.green, gdk_rgba.blue, gdk_rgba.alpha]

def utilities_color_array_to_gdk_rgba(r, g, b, a):
	"""Convert an array of values (as used by cairo) to a Gdk.RGBA object. The
	values should be integers between 0 and 255, except alpha."""
	if a > 1.0:
		raise Exception("Incorrect value for alpha")
	return Gdk.RGBA(red=r / 255, green=g / 255, blue=b / 255, alpha=a)

################################################################################

def utilities_rgba_to_hexadecimal(r, g, b, a):
	"""Methods such as `GdkPixbuf.Pixbuf.fill` wants an hexadecimal integer
	whose format is 0xrrggbbaa so here are ugly binary operators.
	All four values (even alpha!) must be between 0 and 255."""
	r, g, b, a = int(r), int(g), int(b), int(a)
	return (((((r << 8) + g) << 8) + b) << 8) + a

def utilities_rgb_to_hexadecimal(r, g, b):
	"""Methods such as `GdkPixbuf.Pixbuf.composite_color_simple` want an
	hexadecimal integer whose format is 0xaarrggbb so here are ugly binary
	operators."""
	return (r << 16) + (g << 8) + b

def utilities_gdk_rgba_to_hexadecimal(gdk_rgba):
	"""Returns a displayable string of the hexadecimal code corresponding to a
	Gdk.RGBA object."""
	r = int(255 * gdk_rgba.red)
	g = int(255 * gdk_rgba.green)
	b = int(255 * gdk_rgba.blue)
	hexa_raw = hex(utilities_rgb_to_hexadecimal(r, g, b))
	hexa_string = str(hexa_raw)[2:]
	hexa_string = "0" * (6 - len(hexa_string)) + hexa_string
	return "#" + hexa_string

################################################################################

def utilities_gdk_rgba_from_xy(surface, event_x, event_y):
	rgba_vals = utilities_get_rgba_for_xy(surface, event_x, event_y)
	if rgba_vals is None:
		return # event outside of the surface
	rgba_vals = [*rgba_vals]
	rgba_vals[3] /= 255 # alpha has to be between 0 and 1
	return utilities_color_array_to_gdk_rgba(*rgba_vals)

################################################################################

def utilities_get_rgba_name(gdk_rgba):
	"""To improve accessibility, it is useful to display the name of the colors.
	Sadly, it's a mess to implement, and it's quite approximative."""
	[red, green, blue, alpha] = utilities_gdk_rgba_to_normalized_array(gdk_rgba)
	color_string = ""
	alpha_string = ""
	if alpha == 0.0:
		return _("Transparent")
	elif alpha < 1.0:
		alpha_string = " - " + _("%s%% transparent") % int(100 - alpha * 100)

	total = red + green + blue
	orange_coef = 0.0
	lumin = total / 3.0
	# print(lumin)
	if green != 0:
		orange_coef = (red/green) * lumin

	if total != 0:
		rgb_percents = [red/total, green/total, blue/total]
	else:
		rgb_percents = [0.333, 0.333, 0.333]
	# print(rgb_percents)

	grey_coef_r = rgb_percents[0] * lumin / 3
	grey_coef_g = rgb_percents[1] * lumin / 3
	grey_coef_b = rgb_percents[2] * lumin / 3
	is_grey = abs(grey_coef_r - grey_coef_g) < 0.01
	is_grey = is_grey and abs(grey_coef_g - grey_coef_b) < 0.01
	is_grey = is_grey and abs(grey_coef_b - grey_coef_r) < 0.01

	if is_grey:
		if lumin > 0.9:
			color_string = _("White")
		elif lumin < 0.1:
			color_string = _("Black")
		else:
			color_string = _("Grey")

	elif rgb_percents[0] > 0.5 and rgb_percents[1] > 0.2 and rgb_percents[1] < 0.4:
		if orange_coef > 0.87:
			color_string = _("Orange")
		else:
			color_string = _("Brown")

	elif rgb_percents[0] > 0.4 and rgb_percents[1] < 0.3 and rgb_percents[2] < 0.3:
		if lumin < 0.7 and rgb_percents[0] < 0.7:
			# Context: the name of the current color is provided as a tooltip to
			# help users with color blindness, but some color names don't have a
			# clear definition. Here, the app thinks it's probably brown.
			color_string = _("Probably brown")
		else:
			color_string = _("Red")
	elif rgb_percents[1] > 0.4 and rgb_percents[0] < 0.4 and rgb_percents[2] < 0.4:
		color_string = _("Green")
	elif rgb_percents[2] > 0.4 and rgb_percents[0] < 0.3 and rgb_percents[1] < 0.4:
		color_string = _("Blue")

	elif rgb_percents[0] > 0.3 and rgb_percents[1] > 0.3 and rgb_percents[2] < 0.3:
		if rgb_percents[1] < 0.4:
			color_string = _("Probably brown")
		else:
			color_string = _("Yellow")
	elif rgb_percents[0] > 0.3 and rgb_percents[2] > 0.3 and rgb_percents[1] < 0.3:
		if lumin > 0.6 and rgb_percents[1] < 0.1:
			color_string = _("Magenta")
		else:
			color_string = _("Purple")
	elif rgb_percents[1] > 0.3 and rgb_percents[2] > 0.3 and rgb_percents[0] < 0.2:
		if lumin > 0.7:
			color_string = _("Cyan")
		else:
			# Context: the name of the current color is provided as a tooltip to
			# help users with color blindness, but some color names don't have a
			# clear definition. Here, the app thinks it's probably teal.
			# You can translate "teal" with the name of approaching color, like
			# turquoise or green-blue.
			color_string = _("Probably teal")

	else:
		# Context: the name of the current color is provided as a tooltip to
		# help users with color blindness, but some color names don't have a
		# clear definition. Here, the app can't find a corresponding color name.
		color_string = _("Unknown color name")

	# print(color_string)
	return (color_string + alpha_string)

################################################################################




# Licensed under GPL3 https://github.com/maoschanz/drawing/blob/master/LICENSE

from gi.repository import Gdk
from .abstract_filter import AbstractFilter
from .utilities_blur import utilities_blur_surface, BlurType, BlurDirection

class FilterBlur(AbstractFilter):
	__gtype_name__ = 'FilterBlur'

	def __init__(self, filter_id, filters_tool, *args):
		super().__init__(filter_id, filters_tool)
		self._blur_direction = BlurDirection.INVALID
		self._label, self._spinbtn = self._tool.bar.add_spinbtn( \
		                        _("Blur radius"), [5, 1, 99, 1, 10, 0], 2, 'px')
		# it's [value, lower, upper, step_increment, page_increment, page_size]

	def get_preferred_minimum_width(self):
		return self._label.get_preferred_width()[0] + \
		     self._spinbtn.get_preferred_width()[0]

	def set_filter_compact(self, is_active, is_compact):
		self._label.set_visible(is_active and not is_compact)
		self._spinbtn.set_visible(is_active)

	def set_attributes_values(self):
		state_as_string = self._tool.get_option_value('filters_blur_dir')
		if state_as_string == 'none':
			self._blur_direction = BlurDirection.BOTH
		elif state_as_string == 'horizontal':
			self._blur_direction = BlurDirection.HORIZONTAL
		elif state_as_string == 'vertical':
			self._blur_direction = BlurDirection.VERTICAL
		else:
			self._blur_direction = BlurDirection.INVALID

	def build_filter_op(self):
		options = {
			'blur_algo': self._tool.blur_algo,
			'blur_direction': self._blur_direction,
			'radius': self._spinbtn.get_value_as_int()
		}
		return options

	def do_filter_operation(self, source_pixbuf, operation):
		blur_algo = operation['blur_algo']
		if blur_algo == BlurType.INVALID:
			return
		b_radius = operation['radius']
		b_direction = operation['blur_direction']

		surface = Gdk.cairo_surface_create_from_pixbuf(source_pixbuf, 0, None)
		scale = self._tool.scale_factor()
		surface.set_device_scale(scale, scale)

		bs = utilities_blur_surface(surface, b_radius, blur_algo, b_direction)
		bp = Gdk.pixbuf_get_from_surface(bs, 0, 0, bs.get_width(), bs.get_height())
		self._tool.get_image().set_temp_pixbuf(bp)

	############################################################################
################################################################################




# select_free.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import cairo
from .abstract_select import AbstractSelectionTool

class ToolFreeSelect(AbstractSelectionTool):
	__gtype_name__ = 'ToolFreeSelect'

	def __init__(self, window, **kwargs):
		# Context: this is a tool to select an area according to a shape that
		# can be freely defined by the user.
		super().__init__('free_select', _("Free selection"), 'tool-select-free-symbolic', window)
		self.closing_precision = 10
		self.closing_x = 0.0
		self.closing_y = 0.0
		self.add_tool_action_simple('selection_close', self._force_close_shape)
		self.set_action_sensitivity('selection_close', False)

	def on_tool_unselected(self, *args):
		super().on_tool_unselected()
		self.set_action_sensitivity('selection_close', False)

	def get_editing_tips(self):
		tips = super().get_editing_tips()
		future_path = self.get_selection().get_future_path()
		if future_path is not None and not self.selection_is_active():
			label_instruction = self.label + " - " + \
			                  _("Click on the shape's first point to close it.")
			tips.append(label_instruction)
		return tips

	############################################################################

	def press_define(self, event_x, event_y):
		self._draw_shape(event_x, event_y)
		self.set_action_sensitivity('selection_close', True)

	def motion_define(self, event_x, event_y, render):
		self._draw_shape(event_x, event_y, render)

	def release_define(self, surface, event_x, event_y):
		if self._draw_shape(event_x, event_y):
			self.restore_pixbuf()
			self.operation_type = 'op-define'
			operation = self.build_operation()
			self.apply_operation(operation)
			self.set_action_sensitivity('selection_close', False)
		else:
			return # without updating the surface so the path is visible

	############################################################################

	def _force_close_shape(self, *args):
		self.release_define(None, self.closing_x, self.closing_y)

	def _draw_shape(self, event_x, event_y, render=True):
		"""This method is specific to the 'free selection' mode. It returns a
		boolean, true if the shape should be closed."""
		cairo_context = self.get_context()
		cairo_context.set_source_rgba(0.5, 0.5, 0.5, 0.5)
		thickness = self.get_overlay_thickness()
		cairo_context.set_dash([3 * thickness, 3 * thickness])
		cairo_context.set_line_width(thickness)
		if self.get_selection().get_future_path() is None:
			self.closing_x = event_x
			self.closing_y = event_y
			cairo_context.move_to(event_x, event_y)
			self._pre_load_path(cairo_context.copy_path())
			return False
		delta_x = max(event_x, self.closing_x) - min(event_x, self.closing_x)
		delta_y = max(event_y, self.closing_y) - min(event_y, self.closing_y)
		cairo_context.append_path(self.get_selection().get_future_path())
		if (delta_x < self.closing_precision) and (delta_y < self.closing_precision):
			cairo_context.close_path()
			cairo_context.stroke_preserve()
			self._pre_load_path(cairo_context.copy_path())
			return True
		else:
			cairo_context.line_to(event_x, event_y)
			cairo_context.stroke_preserve() # draw the line without closing the path
			self._pre_load_path(cairo_context.copy_path())
			if render:
				self.non_destructive_show_modif()
			return False

	############################################################################
################################################################################




# preferences.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Gio, Gdk
from .utilities_units import utilities_add_unit_to_spinbtn

@Gtk.Template(resource_path='/com/github/maoschanz/drawing/ui/preferences.ui')
class DrPrefsWindow(Gtk.Window):
	__gtype_name__ = 'DrPrefsWindow'

	content_area = Gtk.Template.Child()
	stack = Gtk.Template.Child()

	page_images = Gtk.Template.Child()
	page_tools = Gtk.Template.Child()
	page_advanced = Gtk.Template.Child()

	adj_width = Gtk.Template.Child()
	adj_height = Gtk.Template.Child()
	adj_preview = Gtk.Template.Child()

	_current_grid = None
	_grid_attach_cpt = 0
	_gsettings = Gio.Settings.new('com.github.maoschanz.drawing')

	def __init__(self, is_beta, wants_csd, **kwargs):
		super().__init__(**kwargs)
		if wants_csd:
			header_bar = Gtk.HeaderBar(
				visible=True, \
				title=_("Preferences"), \
				show_close_button=True \
			)
			self.set_titlebar(header_bar)
			stack_switcher = Gtk.StackSwitcher(
				visible=True, \
				stack=self.stack, \
				halign=Gtk.Align.CENTER \
			)
			header_bar.set_custom_title(stack_switcher)
			self.set_default_size(480, 500)
		else:
			stack_sidebar = Gtk.StackSidebar(visible=True, stack=self.stack)
			stack_sidebar.set_size_request(140, -1)
			self.content_area.pack_start(stack_sidebar, False, False, 0)
			self.set_default_size(600, 400) # Not high enough but the golden
			# ratio is more important than usability

		if is_beta:
			self.get_style_context().add_class('devel')

		self.page_builder_images()
		self.page_builder_tools()
		self.page_builder_advanced(is_beta)

	# Each `page_*` attribute is a GtkGrid. Each `page_builder_*` method declare
	# its grid to be the currently filled one, and reset the counter.
	# Then, each `page_builder_*` method will call the `add_*` methods, who will
	# build accurate widgets to be packed on the grid by the `attach_*` methods.

	############################################################################

	def set_current_grid(self, grid):
		self._current_grid = grid
		self._grid_attach_cpt = 0

	def update_grid_cpt(self):
		self._grid_attach_cpt = self._grid_attach_cpt + 1

	############################################################################

	def page_builder_images(self):
		"""Adds the widget to the grid of the 'images' page."""
		self.set_current_grid(self.page_images)

		# Context: title of a section of the preferences
		self.add_section_title(_("New images"))
		self.add_adj(_("Default width"), 'default-width', self.adj_width)
		self.add_adj(_("Default height"), 'default-height', self.adj_height)
		self.add_colorbtn(_("Default color"), 'default-rgba')

		self.add_section_separator()
		# Context: title of a section of the preferences
		self.add_section_title(_("Images saving"))
		self.add_help(_("JPEG and BMP images can't handle transparency.") + \
		        " " + _("If you save your images in these formats, what " + \
		                  "do want to use to replace transparent pixels?"))
		alpha_dict = {
			'initial': _("Default color"),
			'white': _("White"),
			'black': _("Black"),
			'checkboard': _("Checkboard"),
			'nothing': _("Nothing"),
			'ask': _("Ask before saving")
		}
		# 'nothing' is intentionally removed from this dict, because it would
		# make the 'save_alphaless' action ineffective on PNG files.
		self.add_radio_flowbox('replace-alpha', alpha_dict)

		self.add_section_separator()
		self.add_section_title(_("Zoom"))
		zoom_help = _("You can zoom with Ctrl+scrolling, or only scrolling.") + \
		                   " " + _("See the user help manual for explanations.")
		self.add_help(zoom_help)
		self.add_switch(_("Use 'Ctrl' to zoom"), 'ctrl-zoom')

	def page_builder_tools(self):
		"""Adds the widget to the grid of the 'tools' page."""
		self.set_current_grid(self.page_tools)

		# Context: title of a section of the preferences (appearance of the
		# tools: big icons?, labels?)
		self.add_section_title(_("Appearance"))
		self.add_switch(_("Show tools names"), 'show-labels')
		self.add_switch(_("Use big icons"), 'big-icons')

		self.add_section_separator()
		# Context: title of a section of the preferences
		self.add_section_title(_("Additional tools"))
		self.add_help(_("These tools are not as reliable and useful as " + \
		       "they should be, so they are not all enabled by default."))
		tools_dict = {
			'eraser': _("Eraser"),
			'highlight': _("Highlighter"),
			'brush': _("Brush"),
			'points': _("Points"),
			'free_select': _("Free selection"),
			'color_select': _("Color selection"),
			'picker': _("Color Picker"),
			'paint': _("Paint"),
			'skew': _("Skew"),
		}
		self.add_check_flowbox('disabled-tools', tools_dict)

	def page_builder_advanced(self, is_beta):
		"""Adds the widget to the grid of the 'advanced' page."""
		self.set_current_grid(self.page_advanced)

		# Context: title of a section of the preferences
		self.add_section_title(_("Advanced options"))
		self.add_adj(_("Preview size"), 'preview-size', self.adj_preview)
		if is_beta:
			# This label will not be displayed in the UI of stable versions
			self.add_switch(_("Development features"), 'devel-only')
		else:
			self._gsettings.set_boolean('devel-only', False)
		self.add_colorbtn(_("Background color"), 'ui-background-rgba')

		self.add_switch(_("Prefer dark theme variant"), 'dark-theme-variant')

		self.add_section_separator()
		# Context: title of a section of the preferences. It corresponds to the
		# window layout (header-bar? tool-bar? menu-bar?)
		self.add_section_title(_("Layout"))
		self.add_help(_("The recommended value is \"Automatic\"."))
		# h = headerbar; g = gnome; e = elementary
		# m = menubar; t = toolbar
		# c = color; s = symbolic
		layouts_dict = {
			# It has to match what's written in the previous string.
			'': _("Automatic"),
			'hg': _("Compact"),
			'he': _("elementary OS"),
			# "Legacy" is about the window layout, it means menubar+toolbar, you
			# can translate it like if it was "Traditional"
			'mtc': _("Legacy"),
			# "Legacy" is about the window layout, it means menubar+toolbar, you
			# can translate it like if it was "Traditional".
			# Symbolic icons are monochrome icons.
			'mts': _("Legacy (symbolic icons)"),
			'm': _("Menubar only"),
			'tc': _("Toolbar only"),
			# Symbolic icons are monochrome icons.
			'ts': _("Toolbar only (symbolic icons)")
		}
		self.add_radio_flowbox('deco-type', layouts_dict)

	############################################################################
	# Generic methods to build and pack widgets ################################

	def add_section_separator(self):
		self.attach_large(Gtk.Separator())

	def add_section_title(self, label_text):
		label = Gtk.Label(halign=Gtk.Align.START, use_markup=True, \
		                                    label=('<b>' + label_text + '</b>'))
		self.attach_large(label)

	def add_help(self, label_text):
		help_btn = Gtk.Button.new_from_icon_name('help-faq-symbolic', \
		                                                    Gtk.IconSize.BUTTON)
		help_btn.set_valign(Gtk.Align.CENTER)
		help_btn.set_relief(Gtk.ReliefStyle.NONE)
		help_btn.set_action_name('app.help_prefs') # could be a parameter

		label = Gtk.Label(label=label_text)
		label.set_line_wrap(True)
		label.get_style_context().add_class('dim-label')

		box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=10)
		box.pack_start(label, expand=False, fill=False, padding=0)
		box.pack_end(help_btn, expand=False, fill=False, padding=0)
		box.show_all()
		self.attach_large(box)

	def add_row(self, label_text, widget):
		label = Gtk.Label(label=label_text)
		self.attach_two(label, widget)

	def add_switch(self, label_text, key):
		switch = Gtk.Switch()
		switch.set_active(self._gsettings.get_boolean(key))
		switch.connect('notify::active', self.on_bool_changed, key)
		self.add_row(label_text, switch)

	def add_colorbtn(self, label_text, key):
		color_btn = Gtk.ColorButton(use_alpha=True)
		background_rgba = self._gsettings.get_strv(key)
		r = float(background_rgba[0])
		g = float(background_rgba[1])
		b = float(background_rgba[2])
		a = float(background_rgba[3])
		color = Gdk.RGBA(red=r, green=g, blue=b, alpha=a)
		color_btn.set_rgba(color)
		color_btn.connect('color-set', self.on_colorbtn_changed, key)
		self.add_row(label_text, color_btn)

	def add_adj(self, label_text, key, adj):
		spinbtn = Gtk.SpinButton(adjustment=adj)
		spinbtn.set_value(self._gsettings.get_int(key))
		utilities_add_unit_to_spinbtn(spinbtn, 4, 'px')
		spinbtn.connect('value-changed', self.on_adj_changed, key)
		self.add_row(label_text, spinbtn)

	def add_radio_flowbox(self, setting_key, labels_dict):
		flowbox = Gtk.FlowBox(selection_mode=Gtk.SelectionMode.NONE, expand=True)
		self._radio_are_active = False
		w0 = None
		for id0 in labels_dict:
			w0 = self.build_radio_btn(labels_dict[id0], id0, setting_key, w0)
			flowbox.add(w0)
		self._radio_are_active = True
		self.attach_large(flowbox)

	def build_radio_btn(self, label, btn_id, key, group):
		btn = Gtk.RadioButton(label=label, visible=True, group=group)
		active_id = self._gsettings.get_string(key)
		btn.set_active(btn_id == active_id)
		btn.connect('toggled', self.on_radio_btn_changed, key, btn_id)
		return btn

	def add_check_flowbox(self, setting_key, labels_dict):
		flowbox = Gtk.FlowBox(selection_mode=Gtk.SelectionMode.NONE, expand=True)
		for id0 in labels_dict:
			w0 = self.build_check_btn(labels_dict[id0], id0, setting_key)
			flowbox.add(w0)
		self.attach_large(flowbox)

	def build_check_btn(self, label, row_id, key):
		btn = Gtk.CheckButton(label=label, visible=True)
		array_of_strings = self._gsettings.get_strv(key)
		btn.set_active(row_id not in array_of_strings)
		btn.connect('toggled', self.on_check_btn_changed, key, row_id)
		return btn

	############################################################################
	# Generic callbacks ########################################################

	def on_bool_changed(self, switch, state, key):
		self._gsettings.set_boolean(key, switch.get_active())

	def on_adj_changed(self, spinbtn, key):
		self._gsettings.set_int(key, spinbtn.get_value_as_int())

	def on_check_btn_changed(self, checkbtn, key, btn_id):
		array_of_strings = self._gsettings.get_strv(key)
		if checkbtn.get_active():
			array_of_strings.remove(btn_id)
		else:
			array_of_strings.append(btn_id)
		self._gsettings.set_strv(key, array_of_strings)

	def on_radio_btn_changed(self, radiobtn, key, btn_id):
		if self._radio_are_active:
			self._gsettings.set_string(key, btn_id)

	def on_colorbtn_changed(self, color_btn, key):
		c = color_btn.get_rgba()
		color_array = [str(c.red), str(c.green), str(c.blue), str(c.alpha)]
		self._gsettings.set_strv(key, color_array)

	############################################################################
	# Low-level packing ########################################################

	def attach_two(self, widget1, widget2):
		widget1.set_halign(Gtk.Align.END)
		widget2.set_halign(Gtk.Align.START)
		widget1.set_visible(True)
		widget2.set_visible(True)
		self._current_grid.attach(widget1, 0, self._grid_attach_cpt, 1, 1)
		self._current_grid.attach(widget2, 1, self._grid_attach_cpt, 1, 1)
		self.update_grid_cpt()

	def attach_large(self, widget):
		widget.set_visible(True)
		self._current_grid.attach(widget, 0, self._grid_attach_cpt, 2, 1)
		self.update_grid_cpt()

	############################################################################
################################################################################




# tool_shape.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import cairo, math
from .abstract_classic_tool import AbstractClassicTool
from .utilities_paths import utilities_smooth_path

class ToolShape(AbstractClassicTool):
	__gtype_name__ = 'ToolShape'

	def __init__(self, window, **kwargs):
		super().__init__('shape', _("Shape"), 'tool-freeshape-symbolic', window)
		self.use_operator = True

		self._reset_temp_points()

		self.add_tool_action_simple('shape_close', self._force_close_shape)
		self.set_action_sensitivity('shape_close', False)

		self.load_tool_action_enum('shape_type', 'last-active-shape')
		self._set_active_shape() # that's to initialize a consistent join_id too

		self._filling_id = self.load_tool_action_enum('shape_filling', \
		                                                   'last-shape-filling')

		self._outline_id = 'solid'
		self._outline_label = _("Solid outline")
		self.add_tool_action_enum('shape_outline', self._outline_id)

	def _reset_temp_points(self):
		self._path = None
		self.x_press = None
		self.y_press = None
		self.initial_x = None
		self.initial_y = None

	def get_tooltip(self, event_x, event_y, motion_behavior):
		if motion_behavior != 1:
			return None # no line is being drawn
		if self._shape_id in ['polygon', 'freeshpae']:
			return None # no tooltip for these shapes

		delta_x = abs(self.x_press - event_x)
		delta_y = abs(self.y_press - event_y)

		if self._shape_id == 'circle':
			length = round(math.sqrt(delta_x * delta_x + delta_y * delta_y), 2)
			return _("Radius: %spx") % length

		line1 = _("Width: %spx") % str(delta_x)
		line2 = _("Height: %spx") % str(delta_y)
		return line1 + "\n" + line2

	############################################################################

	def _set_filling_style(self):
		self._filling_id = self.get_option_value('shape_filling')

	def _set_outline_style(self):
		self._outline_id = self.get_option_value('shape_outline')

	def _set_active_shape(self, *args):
		self._shape_id = self.get_option_value('shape_type')
		if self._shape_id == 'rectangle' or self._shape_id == 'polygon':
			self._join_id = cairo.LineJoin.MITER
			# maybe BEVEL for polygon?
		else:
			self._join_id = cairo.LineJoin.ROUND

	def get_options_label(self):
		return _("Shape options")

	def get_editing_tips(self):
		self._set_active_shape()
		shape_name = {
			'rectangle': _("Rectangle"),
			'roundedrect': _("Rounded rectangle"),
			'oval': _("Oval"),
			'circle': _("Circle"),
			'polygon': _("Polygon"),
			'freeshape': _("Free shape"),
		}[self._shape_id]

		label_options = shape_name
		self._set_filling_style()
		self._set_outline_style()
		if self._outline_id != 'solid':
			label_options += " - " + {
				'solid': _("Solid outline"),
				'dashed': _("Dashed outline"),
				'none': _("No outline"),
			}[self._outline_id]
		if self._filling_id != 'empty':
			label_options += " - " + {
				'empty': _("Empty shape"),
				# Context: fill a shape with the color of the left click
				'filled': _("Filled with main color"),
				# Context: fill a shape with the color of the right click
				'secondary': _("Filled with secondary color"),
				'h-gradient': _("Horizontal gradient"),
				'v-gradient': _("Vertical gradient"),
				'r-gradient': _("Radial gradient"),
			}[self._filling_id]

		if (self._shape_id == 'polygon' or self._shape_id == 'freeshape') and \
		                                                 self._path is not None:
			label_instruction = shape_name + " - " + \
			                  _("Click on the shape's first point to close it.")
		else:
			label_instruction = None
			self.set_action_sensitivity('shape_close', False)

		if self.get_image().get_mouse_is_pressed():
			label_modifiers = None
		else:
			label_modifiers = _("Press <Alt>, <Shift>, or both, to quickly " + \
			                                      "change the 'filling' option")

		full_list = [label_options, label_instruction, label_modifiers]
		return list(filter(None, full_list))

	def give_back_control(self, preserve_selection):
		self.restore_pixbuf()
		self._reset_temp_points()

	############################################################################

	def on_press_on_area(self, event, surface, event_x, event_y):
		self.last_mouse_btn = event.button
		self.set_common_values(self.last_mouse_btn, event_x, event_y)

		self.update_modifier_state(event.state)
		if 'SHIFT' in self._modifier_keys and 'ALT' in self._modifier_keys:
			self._filling_id = 'secondary'
		elif 'SHIFT' in self._modifier_keys:
			self._filling_id = 'empty'
			if self._outline_id == 'none':
				self._outline_id = 'solid'
		elif 'ALT' in self._modifier_keys:
			self._filling_id = 'filled'
			self._outline_id = 'none'

	def on_motion_on_area(self, event, surface, event_x, event_y, render=True):
		if self._shape_id == 'freeshape':
			operation = self._add_point(event_x, event_y, True)
		elif self._shape_id == 'polygon':
			operation = self._add_point(event_x, event_y, False)
		else:
			if self._shape_id == 'rectangle':
				self._draw_rectangle(event_x, event_y)
			elif self._shape_id == 'roundedrect':
				self._draw_roundedrect(event_x, event_y)
			elif self._shape_id == 'oval':
				self._draw_ellipse(event_x, event_y)
			elif self._shape_id == 'circle':
				self._draw_circle(event_x, event_y)
			operation = self.build_operation(self._path)
		if render:
			self.do_tool_operation(operation)

	def on_release_on_area(self, event, surface, event_x, event_y):
		if self._shape_id == 'freeshape' or self._shape_id == 'polygon':
			operation = self._add_point(event_x, event_y, True)
			self.set_action_sensitivity('shape_close', not operation['closed'])
			if operation['closed']:
				self.apply_operation(operation)
				self._reset_temp_points()
			else:
				self.do_tool_operation(operation)
				self.non_destructive_show_modif()
		else:
			if self._shape_id == 'rectangle':
				self._draw_rectangle(event_x, event_y)
			elif self._shape_id == 'roundedrect':
				self._draw_roundedrect(event_x, event_y)
			elif self._shape_id == 'oval':
				self._draw_ellipse(event_x, event_y)
			elif self._shape_id == 'circle':
				self._draw_circle(event_x, event_y)
			operation = self.build_operation(self._path)
			self.apply_operation(operation)
			self._reset_temp_points()

	############################################################################

	def _force_close_shape(self, *args):
		self.set_common_values(self.last_mouse_btn, self.x_press, self.y_press)
		self.on_release_on_area(None, None, self.initial_x, self.initial_y)

	def _add_point(self, event_x, event_y, memorize):
		"""Add a point to a shape (used by both freeshape and polygon)."""
		cairo_context = self.get_context()
		if self.initial_x is None:
			# print('init polygon')
			(self.initial_x, self.initial_y) = (self.x_press, self.y_press)
			cairo_context.move_to(self.x_press, self.y_press)
			self._path = cairo_context.copy_path()
		else:
			cairo_context.append_path(self._path)
		should_close = self._should_close_shape(event_x, event_y)
		if not should_close:
			# print('continue polygon')
			cairo_context.line_to(event_x, event_y)
		if memorize:
			# print('memorize polygon')
			self._path = cairo_context.copy_path()
		operation = self.build_operation(cairo_context.copy_path())
		operation['closed'] = should_close
		return operation

	def _should_close_shape(self, event_x, event_y):
		if self.initial_x is None or self.initial_y is None:
			return False
		delta_x = max(event_x, self.initial_x) - min(event_x, self.initial_x)
		delta_y = max(event_y, self.initial_y) - min(event_y, self.initial_y)
		closing_limit = max(2, self.tool_width)
		return (delta_x < closing_limit and delta_y < closing_limit)

	############################################################################

	def _draw_rectangle(self, event_x, event_y):
		cairo_context = self.get_context()
		cairo_context.move_to(self.x_press, self.y_press)
		cairo_context.line_to(self.x_press, event_y)
		cairo_context.line_to(event_x, event_y)
		cairo_context.line_to(event_x, self.y_press)
		cairo_context.close_path()
		self._path = cairo_context.copy_path()

	def _draw_roundedrect(self, event_x, event_y):
		cairo_context = self.get_context()
		a = min(self.x_press, event_x)
		b = max(self.x_press, event_x)
		c = min(self.y_press, event_y)
		d = max(self.y_press, event_y)
		radius = min(d - c, b - a) / 6 # c'est arbitraire
		pi2 = math.pi / 2
		cairo_context.arc(a + radius, c + radius, radius, 2 * pi2, 3 * pi2)
		cairo_context.arc(b - radius, c + radius, radius, 3 * pi2, 4 * pi2)
		cairo_context.arc(b - radius, d - radius, radius, 0 * pi2, 1 * pi2)
		cairo_context.arc(a + radius, d - radius, radius, 1 * pi2, 2 * pi2)
		cairo_context.close_path()
		self._path = cairo_context.copy_path()

	def _draw_ellipse(self, event_x, event_y):
		cairo_context = self.get_context()
		saved_matrix = cairo_context.get_matrix()

		halfw = int((self.x_press - event_x) / 2)
		halfh = int((self.y_press - event_y) / 2)
		# Ensure the matrix will be invertible
		if halfw == 0:
			halfw = 1
		if halfh == 0:
			halfh = 1
		cairo_context.translate(event_x + halfw, event_y + halfh)
		cairo_context.scale(halfw, halfh)
		cairo_context.arc(0, 0, 1, 0, 2 * math.pi)

		cairo_context.set_matrix(saved_matrix)
		cairo_context.close_path()
		self._path = cairo_context.copy_path()

	def _draw_circle(self, event_x, event_y):
		cairo_context = self.get_context()
		delta_x2 = (self.x_press - event_x) * (self.x_press - event_x)
		delta_y2 = (self.y_press - event_y) * (self.y_press - event_y)
		rayon = math.sqrt(delta_x2 + delta_y2)
		cairo_context.arc(self.x_press, self.y_press, rayon, 0.0, 2 * math.pi)
		self._path = cairo_context.copy_path()

	############################################################################

	def build_operation(self, cairo_path):
		pixelart_mode = self.get_image().is_zoomed_surface_sharp()
		operation = {
			'tool_id': self.id,
			'rgba_main': self.main_color,
			'rgba_secd': self.secondary_color,
			'antialias': self._use_antialias,
			'operator': self._operator,
			'line_join': self._join_id,
			'line_width': self.tool_width,
			'filling': self._filling_id,
			'outline': self._outline_id,
			'smooth': (self._shape_id == 'freeshape') and not pixelart_mode,
			'closed': True,
			'path': cairo_path
		}
		return operation

	def get_pattern_h(self, xmin, xmax):
		pattern = cairo.LinearGradient(xmin, 0.0, xmax, 0.0)
		return pattern

	def get_pattern_v(self, ymin, ymax):
		pattern = cairo.LinearGradient(0.0, ymin, 0.0, ymax)
		return pattern

	def get_pattern_r(self, center_x, center_y, rad):
		pattern = cairo.RadialGradient(center_x, center_y, 0.1 * rad, \
		                               center_x, center_y, 0.9 * rad)
		# the 2 centers could be 2 distinct points
		return pattern

	def _fill_pattern(self, cairo_context, pattern, c1, c2):
		"""Fill the shape defined in the context with a gradient from c1 to c2
		according to the given pattern. The colors are normalized arrays."""
		pattern.add_color_stop_rgba(0.1, *c1)
		pattern.add_color_stop_rgba(0.9, *c2)
		cairo_context.set_source(pattern)
		cairo_context.fill_preserve()

	def _fill_plain(self, cairo_context, color):
		"""Fill the shape defined in the context with the color c."""
		cairo_context.set_source_rgba(*color)
		cairo_context.fill_preserve()

	def do_tool_operation(self, operation):
		cairo_context = self.start_tool_operation(operation)

		line_width = operation['line_width']
		cairo_context.set_line_width(line_width)
		cairo_context.set_line_join(operation['line_join'])

		if operation['smooth']:
			utilities_smooth_path(cairo_context, operation['path'])
		else:
			cairo_context.append_path(operation['path'])
		if operation['closed']:
			cairo_context.close_path()

		cairo_context.set_operator(operation['operator'])
		color_main = operation['rgba_main']
		color_secd = operation['rgba_secd']

		filling = operation['filling']
		if filling == 'secondary':
			self._fill_plain(cairo_context, color_secd)
		elif filling == 'filled':
			self._fill_plain(cairo_context, color_main)
		elif filling == 'h-gradient':
			x1, y1, x2, y2 = cairo_context.path_extents()
			pattern = self.get_pattern_h(x1, x2)
			self._fill_pattern(cairo_context, pattern, color_main, color_secd)
		elif filling == 'v-gradient':
			x1, y1, x2, y2 = cairo_context.path_extents()
			pattern = self.get_pattern_v(y1, y2)
			self._fill_pattern(cairo_context, pattern, color_main, color_secd)
		elif filling == 'r-gradient':
			x1, y1, x2, y2 = cairo_context.path_extents()
			ddx = abs(x1 - x2) / 2
			ddy = abs(y1 - y2) / 2
			center_x = min(x1, x2) + ddx
			center_y = min(y1, y2) + ddy
			rad = max(ddx, ddy)
			pattern = self.get_pattern_r(center_x, center_y, rad)
			self._fill_pattern(cairo_context, pattern, color_main, color_secd)
		else: # filling == 'empty':
			pass

		outline = operation['outline']
		cairo_context.set_source_rgba(*color_main)
		if outline == 'dashed':
			cairo_context.set_dash([2 * line_width, 2 * line_width])
		if outline != 'none':
			cairo_context.stroke()

		# Reset the current path (maybe useful if no outline)
		cairo_context.new_path()

	############################################################################
################################################################################




# Licensed under GPL3 https://github.com/maoschanz/drawing/blob/master/LICENSE

import cairo
from gi.repository import Gdk
from .abstract_filter import AbstractFilter

class FilterColors(AbstractFilter):
	__gtype_name__ = 'FilterColors'

	def __init__(self, filter_id, filters_tool, *args):
		super().__init__(filter_id, filters_tool)

	# this filter could be so much more, but what's pertinent?

	def do_filter_operation(self, source_pixbuf, operation):
		surface = Gdk.cairo_surface_create_from_pixbuf(source_pixbuf, 0, None)
		cairo_context = cairo.Context(surface)
		cairo_context.set_operator(cairo.Operator.DIFFERENCE)
		cairo_context.set_source_rgba(1.0, 1.0, 1.0, 1.0)
		cairo_context.paint()
		new_pixbuf = Gdk.pixbuf_get_from_surface(surface, 0, 0, \
		                              surface.get_width(), surface.get_height())
		self._tool.get_image().set_temp_pixbuf(new_pixbuf)

	############################################################################
################################################################################




# tool_points.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import cairo, math
from .abstract_classic_tool import AbstractClassicTool

class ToolPoints(AbstractClassicTool):
	__gtype_name__ = 'ToolPoints'

	def __init__(self, window, **kwargs):
		# Context: this is the name of a tool to draw little circles, crosses,
		# or squares at precise points of the image, for example to draw a vague
		# graph, or to highlight something in an image.
		# A number can be added on the cross/circle/square to help captioning
		# the elements of an image.
		super().__init__('points', _("Points"), 'tool-points-symbolic', window)

		self.add_tool_action_enum('points_type', 'cross')
		self.add_tool_action_boolean('points_number', False)
		self.add_tool_action_simple('points_reset_num', self._reset_number)
		self.add_tool_action_simple('points_inc_num', self._increment_number)
		self.add_tool_action_simple('points_dec_num', self._decrement_number)

		self._next_number = 1
		self._set_active_type() # Not optimal but more readable

	def _set_active_type(self):
		state_as_string = self.get_option_value('points_type')
		if state_as_string == 'cross':
			self._shape_label = _("Cross")
		elif state_as_string == 'x-cross':
			self._shape_label = _("X-shaped cross")
		elif state_as_string == 'square':
			self._shape_label = _("Square")
		else:
			self._shape_label = _("Circle")
		self._points_type = state_as_string

		self._use_number = self.get_option_value('points_number')
		self.set_action_sensitivity('points_reset_num', self._use_number)
		self.set_action_sensitivity('points_inc_num', self._use_number)
		self.set_action_sensitivity('points_dec_num', self._use_number)

	def _reset_number(self, *args):
		self._next_number = 1
		self.window.set_window_subtitles()

	def _increment_number(self, *args):
		self._next_number += 1
		self.window.set_window_subtitles()

	def _decrement_number(self, *args):
		self._next_number -= 1
		self.window.set_window_subtitles()

	def get_options_label(self):
		return _("Points options")

	def get_editing_tips(self):
		self._set_active_type()
		label = self.label + " - " + self._shape_label
		if self._use_number:
			label += " - " + _("Next number: %s") % self._next_number
		# A single label is returned so the next number is always visible
		return [label]

	############################################################################

	def on_press_on_area(self, event, surface, event_x, event_y):
		self.set_common_values(event.button, event_x, event_y)

	def on_motion_on_area(self, event, surface, event_x, event_y, render=True):
		if render:
			operation = self.build_operation(event_x, event_y)
			self.do_tool_operation(operation)

	def on_release_on_area(self, event, surface, event_x, event_y):
		operation = self.build_operation(event_x, event_y)
		self.apply_operation(operation)
		if self._use_number:
			self._increment_number()

	############################################################################

	def build_operation(self, event_x, event_y):
		if self._use_number:
			number = self._next_number
		else:
			number = None
		operation = {
			'tool_id': self.id,
			'rgba': self.main_color,
			'rgba2': self.secondary_color, # for the font of the number
			'antialias': self._use_antialias,
			'line_width': self.tool_width,
			'point_type': self._points_type,
			'number': number,
			'x': event_x,
			'y': event_y,
		}
		return operation

	def do_tool_operation(self, operation):
		cairo_context = self.start_tool_operation(operation)
		cairo_context.set_line_cap(cairo.LineCap.BUTT)
		cairo_context.set_source_rgba(*operation['rgba'])

		point_width = operation['line_width']
		line_width = max(1, int(point_width / 4))
		half_width = max(1, int(point_width / 2))
		cairo_context.set_line_width(line_width)

		x = operation['x']
		y = operation['y']

		point_type = operation['point_type']
		if point_type == 'circle':
			cairo_context.arc(x, y, half_width, 0.0, 2 * math.pi)
			cairo_context.fill()
		elif point_type == 'cross':
			# Looks awful with small sizes, either with or without antialiasing
			cairo_context.move_to(x, y - half_width)
			cairo_context.line_to(x, y + half_width)
			cairo_context.move_to(x - half_width, y)
			cairo_context.line_to(x + half_width, y)
		elif point_type == 'x-cross':
			cairo_context.move_to(x - half_width, y - half_width)
			cairo_context.line_to(x + half_width, y + half_width)
			cairo_context.move_to(x + half_width, y - half_width)
			cairo_context.line_to(x - half_width, y + half_width)
		else: # if point_type == 'square':
			cairo_context.set_line_width(point_width)
			cairo_context.move_to(x, y - half_width)
			cairo_context.line_to(x, y + half_width)

		cairo_context.stroke() # without operator support, because it wouldn't
		# make much sense, and the abstract method for them doesn't support
		# changing the line width depending on the point type (like here).

		number = operation['number']
		if number is None:
			return

		# Text size
		cairo_context.set_font_size(max(1, int(point_width * 0.8)))

		# Coordinates
		if number < 10:
			num_x = x - line_width
		elif number < 100:
			num_x = x - line_width * 2
		else:
			num_x = x - line_width * 3
		# XXX could be better with cairo.ScaledFont.text_extents()
		num_y = y + line_width

		# Outline
		for dx in [-1, 1]:
			for dy in [-1, 1]:
				cairo_context.move_to(num_x + dx, num_y + dy)
				cairo_context.show_text(str(number))
		# XXX all my text insertions could use a "layer" like the brushes

		# Text color
		cairo_context.set_source_rgba(*operation['rgba2'])

		# Cairo's "toy" text API (enough for numbers)
		cairo_context.move_to(num_x, num_y)
		cairo_context.show_text(str(number))

	############################################################################
################################################################################




# main.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import sys, gi, datetime
gi.require_version('Gtk', '3.0')
gi.require_version('PangoCairo', '1.0')
from gi.repository import Gtk, Gio, GLib, Gdk
from .window import DrWindow
from .preferences import DrPrefsWindow
from .utilities_files import utilities_gfile_is_image

def main(version):
	app = Application(version)
	return app.run(sys.argv)

################################################################################

class Application(Gtk.Application):
	shortcuts_window = None
	prefs_window = None

	APP_ID = 'com.github.maoschanz.drawing'
	APP_PATH = '/com/github/maoschanz/drawing'
	BUG_REPORT_URL = 'https://github.com/maoschanz/drawing/issues/new/choose'
	FLATPAK_BINARY_PATH = '/app/bin/drawing'
	CURRENT_BINARY_PATH = ''

	############################################################################
	# Initialization ###########################################################

	def __init__(self, version):
		super().__init__(application_id=self.APP_ID,
		                 flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE)

		GLib.set_application_name(_("Drawing"))
		GLib.set_prgname(self.APP_ID)
		self._version = version
		self.has_tools_in_menubar = False
		self.runs_in_sandbox = False

		self.connect('startup', self.on_startup)
		self.register(None)
		self.connect('activate', self.on_activate)
		self.connect('command-line', self.on_cli)

		self.add_main_option('version', b'v', GLib.OptionFlags.NONE,
		                     # Description of a command line option
		                     GLib.OptionArg.NONE, _("Show the app version"), None)
		self.add_main_option('new-window', b'n', GLib.OptionFlags.NONE,
		                     # Description of a command line option
		                     GLib.OptionArg.NONE, _("Open a new window"), None)
		self.add_main_option('new-tab', b't', GLib.OptionFlags.NONE,
		                     # Description of a command line option
		                     GLib.OptionArg.NONE, _("Open a new tab"), None)
		self.add_main_option('edit-clipboard', b'c', GLib.OptionFlags.NONE,
		             # Description of a command line option
		             GLib.OptionArg.NONE, _("Edit the clipboard content"), None)

		icon_theme = Gtk.IconTheme.get_default()
		icon_theme.add_resource_path(self.APP_PATH + '/icons')
		icon_theme.add_resource_path(self.APP_PATH + '/tools/icons')

	def on_startup(self, *args):
		"""Called only once, add app-wide menus and actions, and all accels."""
		self._build_actions()
		builder = Gtk.Builder.new_from_resource(self.APP_PATH + '/ui/app-menus.ui')
		menubar_model = builder.get_object('menu-bar')
		self.set_menubar(menubar_model)

	def _build_actions(self):
		"""Add all app-wide actions."""
		self.add_action_simple('new_window', self.on_new_window, ['<Ctrl>n'])
		self.add_action_simple('settings', self.on_prefs)

		current_date = datetime.datetime.now()
		if current_date.month == 4 and current_date.day == 1:
			self.add_action_simple('april-fools', self.on_april_fools)

		self.add_action_simple('help', self.on_help_index, ['F1'])
		self.add_action_simple('help_main', self.on_help_main)
		self.add_action_simple('help_zoom', self.on_help_zoom)
		self.add_action_simple('help_fullscreen', self.on_help_fullscreen)
		self.add_action_simple('help_tools', self.on_help_tools)
		self.add_action_simple('help_colors', self.on_help_colors)
		self.add_action_simple('help_transform', self.on_help_transform)
		self.add_action_simple('help_selection', self.on_help_selection)
		self.add_action_simple('help_prefs', self.on_help_prefs)
		self.add_action_simple('help_whats_new', self.on_help_whats_new)

		self.add_action_simple('report-issue', self.on_report)
		self.add_action_simple('shortcuts', self.on_shortcuts, \
		                                         ['<Ctrl>question', '<Ctrl>F1'])
		self.add_action_simple('about', self.on_about, ['<Shift>F1'])
		self.add_action_simple('quit', self.on_quit, ['<Ctrl>q'])

	############################################################################
	# Opening windows & CLI handling ###########################################

	def open_window_with_content(self, gfile, get_cb, check_duplicates=True):
		"""Open a new window with an optional Gio.File as an argument. If get_cb
		is true, the Gio.File is ignored and the picture is built from the
		clipboard content."""
		if gfile is not None and check_duplicates:
			w, already_opened_index = self.has_image_opened(gfile.get_path())
			if w is not None:
				if not w.confirm_open_twice(gfile):
					w.notebook.set_current_page(already_opened_index)
					return

		win = DrWindow(application=self)
		win.present()

		content_params = {'gfile': gfile, 'get_cb': get_cb}
		# Parameters are: time in milliseconds, method, data # XXX todo?
		# GLib.timeout_add(10, win.init_window_content_async, content_params)
		win.init_window_content_async(content_params)
		return win

	def on_activate(self, *args):
		"""I don't know if this is ever called from the 'activate' signal, but
		it's called anyways."""
		win = self.props.active_window
		if not win:
			self.on_new_window()
		else:
			win.present()

	def on_cli(self, gio_app, gio_command_line):
		"""Main handler, managing options and CLI arguments.
		Arguments are a `Gio.Application` and a `Gio.ApplicationCommandLine`."""

		# This is the list of files given by the command line. If there is none,
		# this will be ['/app/bin/drawing'] which has a length of 1.
		arguments = gio_command_line.get_arguments()
		self.CURRENT_BINARY_PATH = arguments[0]
		if self.CURRENT_BINARY_PATH == self.FLATPAK_BINARY_PATH:
			self.runs_in_sandbox = True

		# Possible options are 'version', 'edit-clipboard', 'new-tab', and
		# 'new-window', in this order: only one option can be applied, '-ntvc'
		# will be understood as '-v'.
		options = gio_command_line.get_options_dict()

		if options.contains('version'):
			print(_("Drawing") + ' ' + self._version)
			if self.is_beta():
				print(_("This version isn't stable!"))
			print()
			print(_("Report bugs or ideas") + " 👉️ " + self.BUG_REPORT_URL)

		elif options.contains('edit-clipboard'):
			win = self.props.active_window
			if not win:
				self.open_window_with_content(None, True)
			else:
				win.present()
				win.build_image_from_clipboard()

		elif options.contains('new-tab') and len(arguments) == 1:
			# If '-t' but no file given as argument
			win = self.props.active_window
			if not win:
				self.on_new_window()
			else:
				win.present()
				win.build_blank_image()

		elif options.contains('new-window'):
			# it opens one new window per file given as argument, or just one
			# new window if no argument is a valid enough file.
			windows_counter = 0
			for fpath in arguments[1:]:
				f = self._get_valid_file(gio_command_line, fpath)
				# here, f can be a GioFile or a boolean. True would mean the app
				# should open a new blank image.
				if f == False:
					continue
				f = None if f == True else f
				self.open_window_with_content(f, False)
				windows_counter = windows_counter + 1
			if windows_counter == 0:
				self.on_new_window()

		elif len(arguments) == 1:
			self.on_activate()

		else:
			# giving files without '-n' is equivalent to giving files with '-t'
			for fpath in arguments[1:]:
				f = self._get_valid_file(gio_command_line, fpath)
				# here f can be a Gio.File or a boolean: True would mean the app
				# should open a new blank image.
				if f == False:
					continue
				win = self.props.active_window
				if not win:
					f = None if f == True else f
					self.open_window_with_content(f, False)
				else:
					win.present()
					if f == True:
						win.build_blank_image()
					else:
						win.build_new_from_file(gfile=f)

		# I don't even know if i should return something
		return 0

	############################################################################
	# Actions callbacks ########################################################

	def on_new_window(self, *args):
		"""Action callback, opening a new window with an empty canvas."""
		return self.open_window_with_content(None, False)

	def on_report(self, *args):
		"""Action callback, opening a new issue on the github repo."""
		win = self.props.active_window
		Gtk.show_uri_on_window(win, self.BUG_REPORT_URL, Gdk.CURRENT_TIME)

	def on_shortcuts(self, *args):
		"""Action callback, showing the 'shortcuts' dialog."""
		if self.shortcuts_window is not None:
			self.shortcuts_window.destroy()
		builder = Gtk.Builder().new_from_resource(self.APP_PATH + '/ui/shortcuts.ui')
		self.shortcuts_window = builder.get_object('shortcuts-window')
		self.shortcuts_window.present()

	def on_prefs(self, *args):
		"""Action callback, showing the preferences window."""
		if self.prefs_window is not None:
			self.prefs_window.destroy()
		wants_csd = 'h' in self.props.active_window.deco_layout
		self.prefs_window = DrPrefsWindow(self.is_beta(), wants_csd, \
		                                                       application=self)
		self.prefs_window.present()

	def on_april_fools(self, *args):
		"""Action callback, rickrolling the user."""
		Gtk.show_uri_on_window(
			self.props.active_window, \
			'https://www.youtube.com/watch?v=dQw4w9WgXcQ',\
			Gdk.CURRENT_TIME \
		)

	def on_help_index(self, *args):
		"""Action callback, showing the index of user help manual."""
		self._show_help_page('')

	def on_help_main(self, *args):
		self._show_help_page('/main_features')

	def on_help_zoom(self, *args):
		self._show_help_page('/zoom_preview')

	def on_help_fullscreen(self, *args):
		self._show_help_page('/fullscreen')

	def on_help_tools(self, *args):
		self._show_help_page('/tools_classic')

	def on_help_colors(self, *args):
		self._show_help_page('/tools_classic_colors')

	def on_help_transform(self, *args):
		self._show_help_page('/tools_transform')

	def on_help_selection(self, *args):
		self._show_help_page('/tools_selection')

	def on_help_prefs(self, *args):
		self._show_help_page('/preferences')

	def on_help_whats_new(self, *args):
		self._show_help_page('/whats_new')

	def on_about(self, *args):
		"""Action callback, showing the "about" dialog."""
		about_dialog = Gtk.AboutDialog(transient_for=self.props.active_window,
			copyright="© 2018-2023 Romain F. T.",
			authors=["Romain F. T.", "Fábio Colacio", "Alexis Lozano"],
			# To tranlators: "translate" this by a list of your names (one name
			# per line), they will be displayed in the "about" dialog
			translator_credits=_("translator-credits"),
			artists=["Tobias Bernard", "Romain F. T.",
			# To translators: this is credits for the icons, consider that "Art
			# Libre" is proper name
			                       _("GNOME's \"Art Libre\" icon set authors")],
			comments=_("Simple image editor for Linux"),
			license_type=Gtk.License.GPL_3_0,
			logo_icon_name=self.APP_ID, version=str(self._version),
			website='https://maoschanz.github.io/drawing/',
			website_label=_("Official webpage"))
		bug_report_btn = Gtk.LinkButton(halign=Gtk.Align.CENTER, visible=True, \
		               label=_("Report bugs or ideas"), uri=self.BUG_REPORT_URL)
		# about_dialog.get_content_area().add(bug_report_btn) # should i?
		about_dialog.set_icon_name('com.github.maoschanz.drawing')

		about_dialog.run()
		about_dialog.destroy()

	def on_quit(self, *args):
		"""Action callback, quitting the entire app."""
		if self.shortcuts_window is not None:
			self.shortcuts_window.destroy()
		if self.prefs_window is not None:
			self.prefs_window.destroy()

		can_quit = True
		# Try (= ask confirmation) to quit the main window(s)
		main_windows = self.get_windows()
		for w in main_windows:
			if w.on_close():
				# User clicked on "cancel"
				can_quit = False
			else:
				w.close()
				w.destroy()

		# The expected behavior, but now theoretically useless, since closing all
		# appwindows should quit automatically. It's too violent to be left
		# without a guard clause.
		if can_quit:
			self.quit()

	############################################################################
	# Utilities ################################################################

	def is_beta(self):
		"""Tells is the app version is even or odd, odd versions being considered
		as unstable versions. This affects available options and the style of
		the headerbar."""
		return (int(self._version.split('.')[1]) * 5) % 10 == 5

	def get_current_version(self):
		return self._version

	def add_action_simple(self, action_name, callback, shortcuts=[]):
		action = Gio.SimpleAction.new(action_name, None)
		action.connect('activate', callback)
		self.add_action(action)
		self.set_accels_for_action('app.' + action_name, shortcuts)

	def add_action_boolean(self, action_name, default, callback):
		action = Gio.SimpleAction().new_stateful(action_name, None, \
		                                      GLib.Variant.new_boolean(default))
		action.connect('change-state', callback)
		self.add_action(action)

	def _show_help_page(self, suffix):
		win = self.props.active_window
		Gtk.show_uri_on_window(win, 'help:drawing' + suffix, Gdk.CURRENT_TIME)

	def has_image_opened(self, file_path):
		"""Returns the window in which the given file is opened, and the index
		of the tab where it is in the window's notebook.
		Or `None, None` otherwise."""
		for win in self.get_windows():
			position_in_window = win.has_image_opened(file_path)
			if position_in_window is not None:
				return win, position_in_window
		return None, None

	def _get_valid_file(self, app, path):
		"""Creates a GioFile object if the path corresponds to an image. If no
		GioFile can be created, it returns a boolean telling whether or not a
		window should be opened anyway."""
		if path == self.CURRENT_BINARY_PATH:
			# when it's CURRENT_BINARY_PATH, the situation is normal (no error)
			# and nothing to open.
			return False

		err = _("Error opening this file.") + ' '
		try:
			gfile = app.create_file_for_arg(path)
		except Exception as excp:
			if self.runs_in_sandbox:
				command = "\n\tflatpak run --file-forwarding {0} @@ {1} @@\n"
				command = command.format(self.APP_ID, path)
				# This is an error message, %s is a better command suggestion
				err = err + _("Did you mean %s ?") % command
			else:
				err = err + excp.message
			print(err) # TODO show that message in an empty window
			return False

		is_image, err = utilities_gfile_is_image(gfile, err)
		if is_image:
			return gfile
		else:
			print(err) # TODO show that message in an empty window
			return True

	############################################################################
################################################################################




# Licensed under GPL3 https://github.com/maoschanz/drawing/blob/master/LICENSE

from gi.repository import Gtk, GLib

class DrDecoManagerMenubar():
	__gtype_name__ = 'DrDecoManagerMenubar'
	UI_PATH = '/com/github/maoschanz/drawing/ui/'
	SUBTITLE_TIME = 3000

	def __init__(self, window, use_menubar):
		self._window = window
		self._main_menu_btn = None

		window.set_show_menubar(use_menubar)
		if use_menubar:
			# that's an arbitrary restriction
			window.set_titlebar(None)

		self._main_title = _("Drawing")
		self._subtitles = [_("Loading…")]
		self._subtitle_index = 0
		self._is_forced = False
		self._loop_is_ongoing = False

	def remove_from_ui(self):
		self._loop_is_ongoing = False
		return False

	############################################################################
	# Management of the title/subtitle(s) ######################################

	def set_title(self, new_title_label):
		if new_title_label == self._main_title:
			return False
		self._main_title = new_title_label
		return True

	def set_subtitles(self, subtitles_list):
		if len(subtitles_list) == len(self._subtitles):
			list_is_similar = True
			for index, subtitle in enumerate(subtitles_list):
				if subtitle != self._subtitles[index]:
					list_is_similar = False
			if list_is_similar:
				return False
		self._subtitles = subtitles_list
		return True

	def force_update_titles(self):
		if self._loop_is_ongoing:
			self._is_forced = True
		else:
			# idiotic hack to start the infinite loop
			self._loop_is_ongoing = True
		self._subtitle_index -= 1
		self._update_titles()

	def _increment_subtitle_index(self):
		"""Instructions needed each 3 seconds, to change the index of the
		subtitle to display, and program the next subtitle change."""
		self._subtitle_index += 1
		if self._subtitle_index <= 0 or self._subtitle_index >= len(self._subtitles):
			self._subtitle_index = 0

		if self._is_forced:
			self._is_forced = False
		elif not self._loop_is_ongoing:
			# The loop has been stopped = the window has been closed
			return
		else:
			GLib.timeout_add(self.SUBTITLE_TIME, self._update_titles, {})

	def _update_titles(self, *args):
		"""This is used as a GSourceFunc so it should return False."""
		self._increment_subtitle_index()

		full_title = self._main_title
		if len(self._subtitles) > 1:
			full_title += ' ~ ' + self._subtitles[self._subtitle_index]
		elif len(self._subtitles) == 1:
			full_title += ' ~ ' + self._subtitles[0]
		self._window.set_title(full_title)
		return False

	############################################################################

	def toggle_menu(self):
		if self._main_menu_btn is not None:
			self._main_menu_btn.set_active(not self._main_menu_btn.get_active())

	def set_undo_label(self, label):
		pass # update "undo" item in the menubar ?

	def set_redo_label(self, label):
		pass # update "redo" item in the menubar ?

	############################################################################
	# Adaptability #############################################################

	def init_adaptability(self):
		pass

	def adapt_to_window_size(self):
		pass

	def set_compact(self, state):
		pass

	############################################################################
################################################################################

class DrDecoManagerHeaderbar(DrDecoManagerMenubar):
	__gtype_name__ = 'DrDecoManagerHeaderbar'

	def __init__(self, is_eos, window):
		super().__init__(window, False)
		self._is_narrow = True # This is reducing the complexity of resizing,
		# but its main goal is to avoid a GTK minor bug where the initial
		# bunch of configure-event signals was sent to soon.

		# Build the window's headerbar. If "is_eos" is true, the headerbar will
		# follow elementaryOS guidelines, else it will follow GNOME guidelines.
		if is_eos:
			resource_path = self.UI_PATH + 'headerbar-eos.ui'
		else:
			resource_path = self.UI_PATH + 'headerbar.ui'
		builder = Gtk.Builder.new_from_resource(resource_path)
		self._widget = builder.get_object('header_bar')
		window.set_titlebar(self._widget)

		# Code differences are kept minimal between the 2 cases: widgets will
		# share similar names in order to both work with the same method
		# updating widgets' visibility when resizing.
		self._save_long = builder.get_object('save_long')
		self._save_short = builder.get_object('save_short')
		self._hidable_widget_1 = builder.get_object('hidable1')
		self._hidable_widget_2 = builder.get_object('hidable2')

		# Mandatory widget name, used for the `win.main_menu` action
		self._main_menu_btn = builder.get_object('main_menu_btn')

		# History buttons whose tooltips depends on the last operation
		self._undo_btn = builder.get_object('undo_btn')
		self._redo_btn = builder.get_object('redo_btn')

		# Quite extreme as a precaution, will be more precise later
		self._limit_size = 750
		self._manual_correction = 0

		builder.add_from_resource(self.UI_PATH + 'win-menus.ui')
		if is_eos:
			self._init_menus_eos(builder)
		else:
			self._init_menus_gnome(builder)

		# The longer one is set by default to be consistent with the initial
		# value of self._is_narrow
		self._main_menu_btn.set_menu_model(self._long_primary_menu)

	def _init_menus_gnome(self, builder):
		"""Sets the menus for the GNOME/Budgie layout: `self._hidable_widget_2`
		is the "New Image" button here."""
		self._short_primary_menu = builder.get_object('short-window-menu')
		self._long_primary_menu = builder.get_object('long-window-menu')
		self._hidable_widget_2.set_menu_model(builder.get_object('new-image-menu'))

	def _init_menus_eos(self, builder):
		"""Sets the menus for the Pantheon layout: the "New Image" button isn't
		hidden here, and menus are shorter."""
		self._short_primary_menu = builder.get_object('minimal-window-menu')
		self._long_primary_menu = builder.get_object('short-window-menu')
		save_as_menubtn = builder.get_object('hidable1')
		save_as_menubtn.set_menu_model(builder.get_object('save-section'))
		share_menubtn = builder.get_object('hidable2')
		share_menubtn.set_menu_model(builder.get_object('share-section'))
		new_btn = builder.get_object('new_btn')
		new_btn.set_menu_model(builder.get_object('new-image-menu'))
		self._manual_correction = 100

	def remove_from_ui(self):
		self._loop_is_ongoing = False
		return self._is_narrow

	############################################################################

	def _update_titles(self, *args):
		if not self._loop_is_ongoing:
			# The loop has been stopped = the window has been closed
			return False
		self._increment_subtitle_index()

		self._widget.set_title(self._main_title)
		if len(self._subtitles) > 1:
			self._widget.set_subtitle(self._subtitles[self._subtitle_index])
		elif len(self._subtitles) == 1:
			self._widget.set_subtitle(self._subtitles[0])
		return False

	def set_undo_label(self, label):
		super().set_undo_label(label)
		if label is None:
			self._undo_btn.set_tooltip_text(_("Undo"))
		else:
			self._undo_btn.set_tooltip_text(_("Undo %s") % label)

	def set_redo_label(self, label):
		super().set_redo_label(label)
		if label is None:
			self._redo_btn.set_tooltip_text(_("Redo"))
		else:
			self._redo_btn.set_tooltip_text(_("Redo %s") % label)

	############################################################################
	# Adaptability #############################################################

	def init_adaptability(self):
		# Header bar width limit
		self._widget.show_all()
		widgets_width = self._hidable_widget_1.get_preferred_width()[0] \
		              + self._hidable_widget_2.get_preferred_width()[0] \
		                     + self._save_long.get_preferred_width()[0] \
		                    - self._save_short.get_preferred_width()[0] \
		                      + self._undo_btn.get_preferred_width()[0] \
		                      + self._redo_btn.get_preferred_width()[0]
		widgets_width = widgets_width + self._manual_correction
		self._limit_size = widgets_width * 2.5 # 100% arbitrary
		# print(self._limit_size)
		self.set_compact(True)
		self.adapt_to_window_size()

	def adapt_to_window_size(self):
		can_expand = (self._widget.get_allocated_width() > self._limit_size)
		incoherent = (can_expand == self._is_narrow)
		if incoherent:
			self.set_compact(not self._is_narrow)

	def set_compact(self, state):
		"""Set the compactness of the headerbar: if the parameter is True, wide
		widgets will be hidden in favor of narrow ones. Else, the opposite."""
		# Instead of a boolean, `state` could be an integer, which would be
		# far more complex to handle, but would allow thinner granularity.
		if state:
			self._main_menu_btn.set_menu_model(self._long_primary_menu)
		else:
			self._main_menu_btn.set_menu_model(self._short_primary_menu)
		self._save_long.set_visible(not state)
		self._save_short.set_visible(state)
		self._hidable_widget_1.set_visible(not state)
		self._hidable_widget_2.set_visible(not state)
		self._is_narrow = state

	############################################################################
################################################################################

class DrDecoManagerToolbar(DrDecoManagerMenubar):
	__gtype_name__ = 'DrDecoManagerToolbar'

	def __init__(self, is_symbolic, with_menubar, window):
		super().__init__(window, with_menubar)
		window.set_titlebar(None) # that's an arbitrary restriction
		if is_symbolic:
			resource_path = self.UI_PATH + 'toolbar-symbolic.ui'
		else:
			resource_path = self.UI_PATH + 'toolbar.ui'
		builder = Gtk.Builder.new_from_resource(resource_path)

		# Composition over inheritance
		self._widget = builder.get_object('toolbar')
		window.toolbar_box.pack_start(self._widget, True, True, 0)
		window.toolbar_box.show_all()

		# Mandatory widget name, used for the `win.main_menu` action
		self._main_menu_btn = builder.get_object('main_menu_btn')

		# History buttons whose tooltips depends on the last operation
		# XXX maybe later
		# self._undo_btn = builder.get_object('undo_btn')
		# self._redo_btn = builder.get_object('redo_btn')

		# The toolbar has menus which need to be set manually
		builder.add_from_resource(self.UI_PATH + 'win-menus.ui')

		new_btn = builder.get_object('new_menu_btn')
		new_menu = Gtk.Menu.new_from_model(builder.get_object('new-image-menu'))
		new_btn.set_menu(new_menu)

		save_btn = builder.get_object('save_menu_btn')
		save_menu = Gtk.Menu.new_from_model(builder.get_object('save-section'))
		save_btn.set_menu(save_menu)

		if with_menubar:
			self._main_menu_btn.set_visible(False)
		else:
			others_menu = builder.get_object('minimal-window-menu')
			self._main_menu_btn.set_menu_model(others_menu)

	def remove_from_ui(self):
		self._loop_is_ongoing = False
		self._widget.destroy()
		return False

	# def set_undo_label(self, label):
	#	super().set_undo_label(label)
	# 	if label is None:
	# 		self._undo_btn.set_tooltip_text(_("Undo"))
	# 	else:
	# 		self._undo_btn.set_tooltip_text(_("Undo %s") % label)

	# def set_redo_label(self, label):
	#	super().set_redo_label(label)
	# 	if label is None:
	# 		self._redo_btn.set_tooltip_text(_("Redo"))
	# 	else:
	# 		self._redo_btn.set_tooltip_text(_("Redo %s") % label)

	############################################################################
################################################################################




# tool_rotate.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import math
from gi.repository import Gdk
from .abstract_transform_tool import AbstractCanvasTool
from .optionsbar_rotate import OptionsBarRotate

class ToolRotate(AbstractCanvasTool):
	__gtype_name__ = 'ToolRotate'

	def __init__(self, window):
		super().__init__('rotate', _("Rotate"), 'tool-rotate-symbolic', window)
		self.cursor_name = 'pointer'
		self.flip_h = False
		self.flip_v = False
		self.angle_press = 0

		self.add_tool_action_simple('rotate-clockwise', self.on_right_clicked)
		self.add_tool_action_simple('rotate-counter-cw', self.on_left_clicked)
		self.add_tool_action_simple('rotate-flip-h', self.on_horizontal_clicked)
		self.add_tool_action_simple('rotate-flip-v', self.on_vertical_clicked)

	def try_build_pane(self):
		self.pane_id = 'rotate'
		self.window.options_manager.try_add_bottom_pane(self.pane_id, self)

	def build_bottom_pane(self):
		pane = OptionsBarRotate(self)
		self.angle_btn = pane.angle_btn
		self.angle_btn.connect('value-changed', self.on_angle_changed)
		return pane

	def get_options_label(self):
		return _("Rotating options")

	def get_editing_tips(self):
		if self.apply_to_selection:
			label_action = _("Rotating the selection")
			label_confirm = None
		else:
			label_action = _("Rotating the canvas")
			label_confirm = self.label + " - " + \
			                         _("Don't forget to confirm the operation!")

		# there is intentionally no `label_modifier_shift` because it would be
		# too hard to explain the actions of each mouse button.

		full_list = [label_action, label_confirm]
		return list(filter(None, full_list))

	def on_tool_selected(self, *args):
		super().on_tool_selected()
		self.flip_h = False
		self.flip_v = False
		self.angle_btn.set_value(0.0)
		self.build_and_do_op() # Show the temp_pixbuf before any event
		if self.apply_to_selection:
			self.cursor_name = 'move'
			# not the ideal cursor, but their is no ideal cursor for this
		else:
			self.cursor_name = 'pointer'
		# the pane is updated by the window according to self.apply_to_selection

	############################################################################

	def on_press_on_area(self, event, surface, event_x, event_y):
		self.update_modifier_state(event.state)
		if 'SHIFT' in self._modifier_keys:
			if event.button == 1:
				self.on_horizontal_clicked()
			elif event.button == 3:
				self.on_vertical_clicked()
			return
		elif not self.apply_to_selection:
			if event.button == 1:
				self.on_left_clicked()
			elif event.button == 3:
				self.on_right_clicked()
			return

		center_x, center_y = self.get_selection().get_center_coords()
		delta_x0 = center_x - event_x
		delta_y0 = center_y - event_y
		press_as_degrees = (math.atan2(delta_x0, delta_y0) * 180) / math.pi
		self.angle_press = self.get_angle() - int(press_as_degrees)

	def on_motion_on_area(self, event, surface, event_x, event_y, render=True):
		if not self.apply_to_selection:
			return
		center_x, center_y = self.get_selection().get_center_coords()
		delta_x = center_x - event_x
		delta_y = center_y - event_y
		release_angle = ( math.atan2(delta_x, delta_y) * 180 ) / math.pi
		self.angle_btn.set_value(int(release_angle) + self.angle_press)
		if render:
			operation = self.build_operation()
			self.do_tool_operation(operation)

	def on_release_on_area(self, event, surface, event_x, event_y):
		self.on_motion_on_area(event, surface, event_x, event_y)

	############################################################################

	def get_angle(self):
		return self.angle_btn.get_value_as_int()

	def on_right_clicked(self, *args):
		angle = self.get_normalized_angle()
		self.angle_btn.set_value(angle - 90)

	def on_left_clicked(self, *args):
		angle = self.get_normalized_angle()
		self.angle_btn.set_value(angle + 90)

	def on_vertical_clicked(self, *args):
		self.flip_v = not self.flip_v
		self.build_and_do_op()

	def on_horizontal_clicked(self, *args):
		self.flip_h = not self.flip_h
		self.build_and_do_op()

	def get_normalized_angle(self, *args):
		angle = self.get_angle() % 360
		angle = int(angle/90) * 90
		return angle

	def on_angle_changed(self, *args):
		if self.get_angle() == 360 or self.get_angle() == -360:
			self.angle_btn.set_value(0)
		self.build_and_do_op()

	############################################################################

	def on_draw_above(self, area, cairo_context):
		x1 = 0
		y1 = 0
		x2 = x1 + self.get_image().temp_pixbuf.get_width()
		y2 = y1 + self.get_image().temp_pixbuf.get_height()
		x1, x2, y1, y2 = self.get_image().get_corrected_coords(x1, x2, y1, y2, \
		                                         self.apply_to_selection, False)
		self._draw_temp_pixbuf(cairo_context, x1, y1)

	############################################################################

	def build_operation(self):
		operation = {
			'tool_id': self.id,
			'is_selection': self.apply_to_selection,
			'is_preview': True,
			'local_dx': 0,
			'local_dy': 0,
			'angle': self.get_angle(),
			'flip_h': self.flip_h,
			'flip_v': self.flip_v
		}
		return operation

	def do_tool_operation(self, operation):
		self.start_tool_operation(operation)
		angle = operation['angle']
		flip_h = operation['flip_h']
		flip_v = operation['flip_v']
		if operation['is_selection']:
			source_pixbuf = self.get_selection_pixbuf()
		else:
			source_pixbuf = self.get_main_pixbuf()

		if angle < 0:
			angle += 360
		gdk_rotation = int(angle / 90) * 90
		cairo_rotation = angle % 90
		# print('angle:', angle)
		# print('gdk_rotation:', gdk_rotation)
		# print('cairo_rotation:', cairo_rotation)
		new_pixbuf = source_pixbuf

		# Image flipping (horizontal or vertical "mirroring")
		if flip_h:
			new_pixbuf = new_pixbuf.flip(True)
		if flip_v:
			new_pixbuf = new_pixbuf.flip(False)

		# Image rotation, using the method from GdkPixbuf.Pixbuf
		new_pixbuf = new_pixbuf.rotate_simple(gdk_rotation)

		# Image rotation, using methods from cairo.Context (only if needed)
		if cairo_rotation != 0:
			surface0 = Gdk.cairo_surface_create_from_pixbuf(new_pixbuf, 0, None)
			surface0.set_device_scale(self.scale_factor(), self.scale_factor())
			coefs = self._get_rotation_matrix(cairo_rotation, \
			                        surface0.get_width(), surface0.get_height())
			new_surface = self.get_deformed_surface(surface0, coefs)
			new_pixbuf = Gdk.pixbuf_get_from_surface(new_surface, 0, 0, \
			                  new_surface.get_width(), new_surface.get_height())

		self.get_image().set_temp_pixbuf(new_pixbuf)
		self.common_end_operation(operation)

	def _get_rotation_matrix(self, angle, width, height):
		"""Transform an angle (in degrees) to the xx/yx/xy/yy coefs expected by
		cairo. Due to previously performed modifications to the data, the angle
		will be between 0 (excluded) and 90 (excluded)."""
		rad = math.pi * angle / 180

		xx = math.cos(rad)
		xy = math.sin(rad)
		yx = -1 * math.sin(rad)
		yy = math.cos(rad)

		x0 = max(0, height * yx)
		y0 = max(0, width * xy)

		return [xx, yx, xy, yy, x0, y0]

	############################################################################
################################################################################




# Licensed under GPL3 https://github.com/maoschanz/drawing/blob/master/LICENSE

import traceback

from .tool_arc import ToolArc
from .tool_brush import ToolBrush
from .tool_eraser import ToolEraser
from .tool_experiment import ToolExperiment
from .tool_highlight import ToolHighlighter
from .tool_line import ToolLine
from .tool_paint import ToolPaint
from .tool_pencil import ToolPencil
from .tool_picker import ToolPicker
from .tool_points import ToolPoints
from .tool_shape import ToolShape
from .tool_text import ToolText

from .select_rect import ToolRectSelect
from .select_free import ToolFreeSelect
from .select_color import ToolColorSelect

from .tool_crop import ToolCrop
from .tool_filters import ToolFilters
from .tool_rotate import ToolRotate
from .tool_scale import ToolScale
from .tool_skew import ToolSkew

################################################################################

class DrToolsInitializer():

	def __init__(self, window):
		self._window = window
		self._tools = {}

	def load_all_tools(self, dev, disabled_tools):

		# The order might be improvable
		self._load_tool('pencil', ToolPencil, disabled_tools, dev)
		self._load_tool('brush', ToolBrush, disabled_tools, dev)
		self._load_tool('eraser', ToolEraser, disabled_tools, dev)
		self._load_tool('highlight', ToolHighlighter, disabled_tools, dev)
		self._load_tool('text', ToolText, disabled_tools, dev)
		self._load_tool('points', ToolPoints, disabled_tools, dev)
		self._load_tool('rect_select', ToolRectSelect, disabled_tools, dev)
		self._load_tool('free_select', ToolFreeSelect, disabled_tools, dev)
		self._load_tool('line', ToolLine, disabled_tools, dev)
		self._load_tool('arc', ToolArc, disabled_tools, dev)
		self._load_tool('shape', ToolShape, disabled_tools, dev)
		self._load_tool('picker', ToolPicker, disabled_tools, dev)
		self._load_tool('color_select', ToolColorSelect, disabled_tools, dev)
		self._load_tool('paint', ToolPaint, disabled_tools, dev)
		if dev:
			self._load_tool('experiment', ToolExperiment, disabled_tools, dev)
		self._load_tool('crop', ToolCrop, disabled_tools, dev)
		self._load_tool('scale', ToolScale, disabled_tools, dev)
		self._load_tool('rotate', ToolRotate, disabled_tools, dev)
		self._load_tool('skew', ToolSkew, disabled_tools, dev)
		self._load_tool('filters', ToolFilters, disabled_tools, dev)

		self._add_auto_mnemonics()
		return self._tools

	############################################################################

	def _load_tool(self, tool_id, tool_class, disabled_tools, dev):
		"""Given its id and its python class, this method tries to load a tool,
		and show an error message if the tool initialization failed."""
		if tool_id not in disabled_tools:
			try:
				self._tools[tool_id] = tool_class(self._window)
			except Exception as err:
				# Context: an error message
				self._window.reveal_action_report(_("Failed to load tool: %s") % tool_id)
				traceback.print_exc()

	def _add_auto_mnemonics(self):
		# I don't want useful tools lacking a mnemonic accelerator because a
		# useless one "stole" its letters, so the mnemonics are decided in the
		# following order:
		sorted_tools = [
			'pencil',
			'text',
			'rect_select',
			'crop',
			'scale',
			'rotate',
			'shape',
			'filters',
			'line',
			'highlight',
			'arc',
			'picker',
			'brush',
			'free_select',
			'eraser',
			'skew',
			'paint',
			'color_select',
			'points',
			'experiment'
		]
		for tool_id in self._tools:
			if tool_id not in sorted_tools:
				print("Warning: " + tool_id + "will not have a mnemonic")

		underlined_chars = {}
		for tool_id in sorted_tools:
			if tool_id not in self._tools:
				continue
			letter_index = 0
			while(letter_index >= 0):
				if letter_index == len(self._tools[tool_id].label):
					letter_index = -1
					continue
				ith_char = self._tools[tool_id].label[letter_index]
				letter_index += 1

				if ith_char.isalpha() \
				and ith_char.upper() not in underlined_chars.values() \
				and ith_char.lower() not in underlined_chars.values():
					underlined_chars[tool_id] = ith_char
					letter_index = -1

		for tool_id in underlined_chars:
			self._tools[tool_id].set_mnemonics(underlined_chars[tool_id])

	############################################################################
################################################################################




# Licensed under GPL3 https://github.com/maoschanz/drawing/blob/master/LICENSE

import cairo
from gi.repository import Gdk
from .abstract_filter import AbstractFilter
from .utilities_blur import utilities_blur_surface, BlurType, BlurDirection

class FilterEmboss(AbstractFilter):
	__gtype_name__ = 'FilterEmboss'

	def do_filter_operation(self, source_pixbuf, operation):
		surface = Gdk.cairo_surface_create_from_pixbuf(source_pixbuf, 0, None)
		scale = self._tool.scale_factor()
		surface.set_device_scale(scale, scale)
		width = source_pixbuf.get_width()
		height = source_pixbuf.get_height()
		new_surface = cairo.ImageSurface(cairo.Format.ARGB32, width, height)

		cairo_context = cairo.Context(new_surface)
		cairo_context.set_source_surface(surface)
		cairo_context.set_operator(cairo.Operator.SOURCE)
		cairo_context.paint()

		bdir = BlurDirection.BOTH # could be an option
		bs = utilities_blur_surface(surface, 1, BlurType.CAIRO_REPAINTS, bdir)
		cairo_context2 = cairo.Context(bs)
		cairo_context2.set_operator(cairo.Operator.DIFFERENCE)
		cairo_context2.set_source_rgba(1.0, 1.0, 1.0, 1.0)
		cairo_context2.paint()

		cairo_context.set_source_surface(bs)
		cairo_context.set_operator(cairo.Operator.OVER)
		cairo_context.paint_with_alpha(0.5)

		new_pixbuf = Gdk.pixbuf_get_from_surface(new_surface, 0, 0, \
		                      new_surface.get_width(), new_surface.get_height())
		self._tool.get_image().set_temp_pixbuf(new_pixbuf)

	############################################################################
################################################################################




# printing_manager.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Gdk
import math

class DrPrintingManager():
	__gtype_name__ = 'DrPrintingManager'

	# Debug purpose only
	_AUTO_SETTINGS = False

	def __init__(self, window):
		self._window = window
		self._auto_orientation = None

	def print_pixbuf(self, pixbuf):
		"""Starts the GTK "print" dialog with a few pre-defined settings."""
		print_op = Gtk.PrintOperation()

		psetup = Gtk.PageSetup()
		if pixbuf.get_height() < pixbuf.get_width():
			self._auto_orientation = Gtk.PageOrientation.LANDSCAPE
		else:
			self._auto_orientation = Gtk.PageOrientation.PORTRAIT
		psetup.set_orientation(self._auto_orientation)
		if self._AUTO_SETTINGS:
			psetup.set_paper_size(Gtk.PaperSize('A4'))
		print_op.set_default_page_setup(psetup)

		# print_op.set_use_full_page(True) # removes margins; do i want it?

		# XXX the default preview doesn't always work, i guess it's because of
		# the sandbox, or the flatpak implementation.
		# The default preview says "evince-previewer" ???
		# I could implement a custom callback to the 'preview' signal but that
		# would be weird. And likely useless since it's only called when the
		# Gtk.PrintOperationAction given to `run` is PREVIEW???
		print_op.connect('draw-page', self._do_draw_page, pixbuf)
		print_op.connect('begin-print', self._do_begin_print, pixbuf)

		if self._AUTO_SETTINGS:
			print_op.set_export_filename("Documents/sortie.pdf")
			action_type = Gtk.PrintOperationAction.EXPORT
		else:
			action_type = Gtk.PrintOperationAction.PRINT_DIALOG
		# `res` will be Gtk.PrintOperationResult.APPLY even when "Preview" is
		# clicked, because the possible values are actually "APPLY", "CANCEL",
		# "IN_PROGRESS" and "ERROR"
		res = print_op.run(action_type, self._window)
		# I'll assume print_op continues to live enough to be used by signals.

	############################################################################

	def _do_draw_page(self, op, print_ctx, page_num, pixbuf):
		self._show_pixbuf_on_page(print_ctx, pixbuf)

	def _do_begin_print(self, op, print_ctx, pixbuf):
		op.set_n_pages(1)
		self._show_pixbuf_on_page(print_ctx, pixbuf)

	# The typical way to use the high-level printing API is to create a
	# Gtk.PrintOperation object with Gtk.PrintOperation.new() when the user
	# selects to print. Then you set some properties on it, e.g. the page size,
	# any Gtk.PrintSettings from previous print operations, the number of pages,
	# the current page, etc.
	#
	# XXX là je ne restaure aucun PrintSettings ce qui est un problème

	def _show_pixbuf_on_page(self, print_ctx, pixbuf):
		"""Paint the (scaled) image on the page using cairo."""
		angle = self._get_delta_orientation(print_ctx)
		use_natural_ratios = (angle == 0 or angle == 180)
		scale = self._get_scale(print_ctx, pixbuf, use_natural_ratios)

		# Rotate the pixbuf
		rotated_pixbuf = pixbuf.rotate_simple(angle)

		# Set the rotated pixbuf as the context's source
		cairo_context = print_ctx.get_cairo_context()
		Gdk.cairo_set_source_pixbuf(cairo_context, rotated_pixbuf, 0, 0)

		# Scale down the context if necessary (otherwise scale may be 1.0)
		cairo_context.scale(scale, scale)

		# XXX the image should be centered in the page

		# Render the pixbuf on the page
		cairo_context.paint()

	def _get_scale(self, print_ctx, pixbuf, use_natural_ratios):
		if use_natural_ratios:
			h_ratio = print_ctx.get_height() / pixbuf.get_height()
			w_ratio = print_ctx.get_width() / pixbuf.get_width()
		else:
			h_ratio = print_ctx.get_height() / pixbuf.get_width()
			w_ratio = print_ctx.get_width() / pixbuf.get_height()

		if h_ratio < 1.0 or w_ratio < 1.0:
			scale = min(h_ratio, w_ratio)
		else:
			scale = 1.0
		return scale

	def _get_delta_orientation(self, print_ctx):
		if self._auto_orientation == Gtk.PageOrientation.PORTRAIT:
			initial_angle = 0
		else: # if self._auto_orientation == Gtk.PageOrientation.LANDSCAPE:
			initial_angle = 90

		current_orientation = print_ctx.get_page_setup().get_orientation()
		if current_orientation == Gtk.PageOrientation.PORTRAIT:
			new_angle = 0
		elif current_orientation == Gtk.PageOrientation.LANDSCAPE:
			new_angle = 90
		elif current_orientation == Gtk.PageOrientation.REVERSE_PORTRAIT:
			new_angle = 180
		else: # if current_orientation == Gtk.PageOrientation.REVERSE_LANDSCAPE:
			new_angle = 270

		delta_angle = new_angle - initial_angle
		if delta_angle < 0:
			delta_angle = delta_angle + 360
		return delta_angle

	############################################################################
################################################################################




# tool_picker.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gdk
from .abstract_classic_tool import AbstractClassicTool
from .utilities_colors import utilities_get_rgba_name, \
                              utilities_gdk_rgba_from_xy, \
                              utilities_gdk_rgba_to_hexadecimal

class ToolPicker(AbstractClassicTool):
	__gtype_name__ = 'ToolPicker'

	def __init__(self, window, **kwargs):
		# Context: this is a tool to pick a RGBA color in the image in order to
		# use it to draw with other tools
		super().__init__('picker', _("Color Picker"), 'color-select-symbolic', window)
		self.use_size = False

	def get_options_model(self):
		return None

	def get_tooltip(self, event_x, event_y, motion_behavior):
		color = utilities_gdk_rgba_from_xy(self.get_surface(), event_x, event_y)
		if color is None:
			return None
		color_name = utilities_get_rgba_name(color)
		color_code = utilities_gdk_rgba_to_hexadecimal(color)
		return color_name + "\n" + color_code

	def on_release_on_area(self, event, surface, event_x, event_y):
		color = utilities_gdk_rgba_from_xy(surface, event_x, event_y)
		if event.button == 1:
			self.window.options_manager.set_left_color(color)
		elif event.button == 3:
			self.window.options_manager.set_right_color(color)
		self.window.back_to_previous()

	############################################################################
################################################################################



# tool_paint.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import cairo
from gi.repository import Gdk
from .abstract_classic_tool import AbstractClassicTool
from .utilities_paths import utilities_get_magic_path
from .utilities_paths import utilities_get_rgba_for_xy

class ToolPaint(AbstractClassicTool):
	__gtype_name__ = 'ToolPaint'

	def __init__(self, window, **kwargs):
		# Context: the name of a tool to fill an area of one color with an other
		super().__init__('paint', _("Paint"), 'tool-paint-symbolic', window)
		self._magic_path = None
		self.use_size = False
		self.add_tool_action_enum('paint_algo', 'replace')

	def get_options_label(self):
		return _("Painting options")

	def get_editing_tips(self):
		paint_algo = self.get_option_value('paint_algo')
		if paint_algo == 'clipping':
			return [_("Click on an area to replace its color by transparency")]
		elif paint_algo == 'whole':
			return [_("Click on the canvas to entirely paint it")]

		label_warning1 = self.label + " - " + _("May not work for complex shapes")
		label_warning2 = self.label + " - " + _("It will not work well if " + \
		                                          "the area's edges are blurry")
		return [label_warning1, label_warning2]

	############################################################################

	def on_press_on_area(self, event, surface, event_x, event_y):
		self.set_common_values(event.button, event_x, event_y)

	def on_release_on_area(self, event, surface, event_x, event_y):
		# Guard clause: we can't paint outside of the surface
		if event_x < 0 or event_x > surface.get_width() \
		or event_y < 0 or event_y > surface.get_height():
			return

		self.old_color = utilities_get_rgba_for_xy(surface, event_x, event_y)

		if self.get_option_value('paint_algo') == 'fill':
			self._magic_path = utilities_get_magic_path(surface, event_x, \
			                                            event_y, self.window, 1)
		elif self.get_option_value('paint_algo') == 'replace':
			self._magic_path = utilities_get_magic_path(surface, event_x, \
			                                            event_y, self.window, 2)
		else:
			pass # == 'clipping' or == 'whole'

		operation = self.build_operation(event_x, event_y)
		self.apply_operation(operation)

	############################################################################

	def build_operation(self, x, y):
		operation = {
			'tool_id': self.id,
			'algo': self.get_option_value('paint_algo'),
			# 'x': x,
			# 'y': y,
			'new_rgba': self.main_color,
			'antialias': self._use_antialias,
			'old_rgba': self.old_color,
			'path': self._magic_path
		}
		return operation

	def do_tool_operation(self, operation):
		self.start_tool_operation(operation) # XXX expose antialiasing option?

		if operation['algo'] == 'replace':
			self._op_replace(operation)
		elif operation['algo'] == 'whole':
			self._op_whole(operation)
		elif operation['algo'] == 'fill':
			self._op_fill(operation)
		else: # == 'clipping'
			self._op_clipping(operation)

	############################################################################

	def _op_whole(self, operation):
		"""Paint the entire image regardless of existing pixels"""
		cairo_context = self.get_context()
		cairo_context.set_source_rgba(*operation['new_rgba'])
		cairo_context.paint()

	def _op_fill(self, operation):
		"""Simple but ugly, and it's relying on the precision of the provided
		path whose creation is based on shitty heurisctics."""
		if operation['path'] is None:
			return
		cairo_context = self.get_context()
		cairo_context.set_source_rgba(*operation['new_rgba'])
		cairo_context.append_path(operation['path'])
		# can_fill = cairo_context.in_fill(operation['x'], operation['y'])
		# print(can_fill)
		# if not can_fill:
		# 	cairo_context.set_fill_rule(cairo.FillRule.EVEN_ODD)
		# 	XXX doesn't work as i expected
		cairo_context.fill()

	def _op_replace(self, operation):
		"""Algorithmically less ugly than `_op_fill`, but the replacing strategy
		itself doesn't handle (semi-) transparent colors correctly. Looks fine
		when there is no alpha at all and the anti-aliasing is off."""
		# In this example, the user paints in blue and clicks on X (red):
		#
		# rrrrrrrrrrrrrrrrrrrrrrrrr
		# rrrgggggggggggggggggrrXrr
		# rrrgaaaaaaaaaaaaaaagrrrrr
		# rrrgaaaaaaaaaaaaaaagrrrrr
		# rrrgggggggggggggggggrraaa
		# rrrrrrrrrrrrrrrrrrrrrraaa
		#
		# All red pixels will be painted blue, but FIXME all the green ones too.
		if operation['path'] is None:
			return

		# First, everything BUT the targeted area is erased
		surface = self.get_surface()
		cairo_context = cairo.Context(surface)
		old_rgba = operation['old_rgba']
		cairo_context.set_source_rgba(255, 255, 255, 1.0)
		cairo_context.append_path(operation['path'])
		cairo_context.set_operator(cairo.Operator.DEST_IN)
		cairo_context.set_line_width(4) # ptêt too much mdr
		cairo_context.set_antialias(cairo.Antialias.NONE)
		cairo_context.stroke_preserve()
		# cairo_context.fill_preserve()

		# The result is loaded as the temp pixbuf…
		self.get_image().set_temp_pixbuf(Gdk.pixbuf_get_from_surface(surface, \
		                       0, 0, surface.get_width(), surface.get_height()))
		# …where the targeted color is replaced by pure alpha
		TOLERANCE = 10 # XXX uuugly
		i = -1 * TOLERANCE
		while i < TOLERANCE:
			red = max(0, old_rgba[0]+i)
			green = max(0, old_rgba[1]+i)
			blue = max(0, old_rgba[2]+i)
			red = int(min(255, red))
			green = int(min(255, green))
			blue = int(min(255, blue))
			self._replace_temp_with_alpha(red, green, blue)
			i = i+1

		# The main pixbuf is restored, since its alterations were a side effect
		# of this highly stupid algorithm
		self.restore_pixbuf()
		# We get a context from it again
		cairo_context2 = self.get_context()

		# The content of the path is cleared
		cairo_context2.append_path(operation['path'])
		cairo_context2.set_operator(cairo.Operator.CLEAR)
		cairo_context2.set_source_rgba(255, 255, 255, 1.0)
		cairo_context2.set_line_width(3) # ptêt too much mdr
		cairo_context.set_antialias(cairo.Antialias.NONE)
		cairo_context2.stroke_preserve()
		cairo_context2.fill()

		# The temp pixbuf (where "only" the pixels around the outline which are
		# not in the targeted color still exist) is shown on top of the image
		cairo_context2.set_operator(cairo.Operator.OVER)
		Gdk.cairo_set_source_pixbuf(cairo_context2, \
		                                     self.get_image().temp_pixbuf, 0, 0)
		cairo_context2.paint()
		self.non_destructive_show_modif()

		# The transparent pixels within the path are painted with the new color
		cairo_context2.set_operator(cairo.Operator.DEST_OVER)
		cairo_context2.set_source_rgba(*operation['new_rgba'])
		cairo_context2.append_path(operation['path'])
		# cairo_context2.paint()
		cairo_context2.set_line_width(3) # ptêt too much mdr
		cairo_context2.stroke_preserve()
		cairo_context2.fill()

	def _op_clipping(self, operation):
		"""Replace the color with transparency by adding an alpha channel."""
		old_rgba = operation['old_rgba']
		r0 = old_rgba[0]
		g0 = old_rgba[1]
		b0 = old_rgba[2]
		# ^ it's not possible to take into account the alpha channel
		margin = 0 # XXX as an option ? is not elegant but it's powerful
		self._clip_red(margin, r0, g0, b0)
		self.restore_pixbuf()
		self.non_destructive_show_modif()

	############################################################################

	def _clip_red(self, margin, r0, g0, b0):
		for i in range(-1 * margin, margin + 1):
			r = r0 + i
			if r <= 255 and r >= 0:
				self._clip_green(margin, r, g0, b0)

	def _clip_green(self, margin, r, g0, b0):
		for i in range(-1 * margin, margin + 1):
			g = g0 + i
			if g <= 255 and g >= 0:
				self._clip_blue(margin, r, g, b0)

	def _clip_blue(self, margin, r, g, b0):
		for i in range(-1 * margin, margin + 1):
			b = b0 + i
			if b <= 255 and b >= 0:
				self._replace_main_with_alpha(r, g, b)

	def _replace_main_with_alpha(self, red, green, blue):
		new_pixbuf = self.get_main_pixbuf().add_alpha(True, red, green, blue)
		self.get_image().set_main_pixbuf(new_pixbuf)

	def _replace_temp_with_alpha(self, red, green, blue):
		new_pixbuf = self.get_image().temp_pixbuf.add_alpha(True, red, green, blue)
		self.get_image().set_temp_pixbuf(new_pixbuf)

	############################################################################
################################################################################



# Licensed under GPL3 https://github.com/maoschanz/drawing/blob/master/LICENSE

import cairo, random
from .abstract_brush import AbstractBrush

class BrushHairy(AbstractBrush):
	__gtype_name__ = 'BrushHairy'

	def _get_tips(self, use_pressure, brush_direction):
		label = _("Hairy brush")
		if use_pressure:
			label += " - " + _("Width depends on the stylus pressure")
		return [label]

	def draw_preview(self, operation, cairo_context):
		cairo_context.set_line_cap(cairo.LineCap.ROUND)
		cairo_context.set_line_join(cairo.LineJoin.BEVEL)
		cairo_context.set_source_rgba(*operation['rgba'])
		super().draw_preview(operation, cairo_context)

	############################################################################

	def do_brush_operation(self, cairo_context, operation):
		"""Straight-shaped brush, like an orientable feather pen. The width is
		pressure-sensitive but otherwise it's NOT speed sensitive."""

		if operation['is_preview']: # Previewing helps performance & debug
			return self.draw_preview(operation, cairo_context)

		self.operation_on_mask(operation, cairo_context)

	def do_masked_brush_op(self, cairo_context, operation):
		cairo_context.set_line_cap(cairo.LineCap.ROUND)
		cairo_context.set_line_join(cairo.LineJoin.ROUND)
		cairo_context.set_line_width(1)

		line_width = max(1, int(operation['line_width'] / 2))

		matrix = []
		random.seed(1) # this hardcoded seed avoids the hairs changing their
		# positions when the user undoes an following operation
		for i in range(line_width * 2):
			matrix.append({
				'dx': random.randint(-1 * line_width, line_width),
				'dy': random.randint(-1 * line_width, line_width)
			})
			# TODO other kinds of loops to define the matrix?

		use_pressure = operation['path'][0]['p'] is not None
		pressure = 1
		for hair in matrix:
			cairo_context.new_path()
			for point in operation['path']:
				if use_pressure and point['p'] > 0:
					pressure = point['p'] * 2
				px = point['x'] + hair['dx'] * pressure
				py = point['y'] + hair['dy'] * pressure
				cairo_context.line_to(px, py)
			cairo_context.stroke()

	############################################################################
################################################################################

